#!/usr/bin/env python
# Encoding: iso-8859-1
# vim: tw=80 ts=4 sw=4 noet
# -----------------------------------------------------------------------------
# Project           :   Kiwi
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre                 <sebastien@type-z.org>
# -----------------------------------------------------------------------------
# Creation date     :   17-Aug-2006
# Last mod.         :   17-Aug-2006
# -----------------------------------------------------------------------------

import re, xml.dom
import sys
from formatting import *
import templates

#------------------------------------------------------------------------------
#
#  Processing functions
#
#------------------------------------------------------------------------------

class Processor(templates.Processor):

	def generate( self, xmlDocument, bodyOnly=False, variables={} ):
		node = xmlDocument.getElementsByTagName("Document")[0]
		self.variables = variables
		if bodyOnly:
			for child in node.childNodes:
				if child.nodeName == "Content":
					return convertContent_bodyonly(child)
		else:
			return convertDocument(node)

#------------------------------------------------------------------------------
#
#  Actual element processing
#
#------------------------------------------------------------------------------

def convertDocument(element):
	return process(element, """\
@Include { doc }
@Document
	@PageHeaders { Simple }
	@FirstPageNumber { 1 }
	@PageOrientation { Portrait }
//
@Text @Begin
$(Header)
$(Header:title)
@BeginSections
$(Content)
@EndSections
@End @Text
""")

def convertContent( element ):
	return process(element, wdiv(element, """<div id='content'>$(*)</div>"""))

def convertContent_bodyonly( element ):
	return process(element, wdiv(element, """$(*)"""))

def convertContent_table( element ):
	return process(element, """<tbody%s>$(*)</tbody>""" % (wattrs(element)))

def convertHeader( element ):
	return process(element, "<title%s>$(Title/title)</title>" % (wattrs(element)))

def convertHeading( element ):
	return process(element, wspan(element, "$(*)"))

def convertSection( element ):
	level = int(element.getAttributeNS(None, "_depth")) + 1
	return process(element,
	  '<div class="section"><a class="link">'
	  + '<h%d class="heading">$(Heading)</h%d></a>' % (level, level)
	  + '<div class="level%d">$(Content:section)</div></div>' % (level)
	)

def convertReferences( element ):
	return process(element, """<div id="references">$(Entry)</div>""")

def convertEntry( element ):
	return process(element, """<div class="entry"><div class="name"><a name="%s">%s</a></div><div class="content">$(*)</div></div>""" %
	(element.getAttributeNS(None, "id"), element.getAttributeNS(None, "id")))

def convertHeader_title( element ):
	return process(element, """<div
	class="title">$(Title/title:header)$(Title/subtitle:header)</div>$(Meta)""")

def converttitle_header( element ):
	return process(element, """<h1%s>$(*)</h1>""" % (wattrs(element)))

def convertsubtitle_header( element ):
	return process(element, """<h2%s>$(*)</h2>""" % (wattrs(element)))

def convertParagraph( element ):
	return process(element, """<p%s>$(*)</p>""" % (wattrs(element)))

def convertParagraph_cell( element ):
	return process(element, """$(*)<br />""")

def convertList( element ):
	return process(element, """<ul%s>$(*)</ul>""" % (wattrs(element)))

def convertListItem( element ):
	return process(element, """<li%s>$(*)</li>""" % (wattrs(element)))

def convertTable( element ):
	return process(element, """<table cellpadding="0" cellspacing="0" align="center">$(Caption)$(Content:table)</table>""")

def convertDefinition( element ):
	return process(element, """<dl%s>$(*)</dl>""" % (wattrs(element)))

def convertDefinitionItem( element ):
	return process(element, """<dt>$(Title)</dt><dd>$(Content)</dd>""")

def convertCaption( element ):
	return process(element, """<caption%s>$(*)</caption>""" % (wattrs(element)))

def convertRow( element ):
	try: index = element.parentNode.childNodes.index(element) % 2 + 1
	except: index = 0 
	classes = ( "", "even", "odd" )
	return process(element, """<tr class='%s'%s>$(*)</tr>""" % (classes[index], wattrs(element)))

def convertCell( element ):
	return process(element, """<td%s>$(*:cell)</td>""" % (wattrs(element)))

def convertBlock( element ):
	title = element.getAttributeNS(None,"title") or element.getAttributeNS(None, "type") or ""
	css_class = ""
	if title:
		css_class=" class='ann%s'" % (element.getAttributeNS(None, "type").capitalize())
		title = "<div class='title'>%s</div>"  % (title.capitalize())
		div_type = "div"
	elif not element.getAttributeNS(None, "type"):
		div_type = "blockquote"
	return process(element, """<%s%s>%s<div class='content'%s>$(*)</div></%s>""" % (div_type, css_class, title, wattrs(element), div_type))

def convertlink( element ):
	if element.getAttributeNS(None, "type") == "ref":
		return process(element, """<a href="#%s">$(*)</a>""" %
		(element.getAttributeNS(None, "target")))
	else:
		# TODO: Support title
		return process(element, """<a href="%s">$(*)</a>""" %
		(element.getAttributeNS(None, "target")))

def convertMeta( element ):
	return process(element, "<table id='meta'>$(*)</table>")

def convertmeta( element ):
	return process(element,
	"<tr><td width='0px' class='name'>%s</td><td width='100%%' class='value'>$(*)</td></tr>" %
	(element.getAttributeNS(None, "name")))

def convertemail( element ):
	mail = ""
	for c in  process(element, """$(*)"""):
		mail += "&#%d;" % (ord(c))
	return """<a href="mailto:%s">%s</a>""" % (mail, mail)

def converturl( element ):
	return process(element, """<a href="$(*)">$(*)</a>""")

def converturl_header( element ):
	return process(element, """<div class='url'>%s</div>""" % (
	converturl(element)))

def convertterm( element ):
	return process(element, """<span class='term'>$(*)</span>""")

def convertquote( element ):
	return process(element, """&ldquo;<span class='quote'>$(*)</span>&rdquo;""")

def convertcitation( element ):
	return process(element, """&laquo;<span class='citation'>$(*)</span>&raquo;""")

def convertemphasis( element ):
	return process(element, """<b>$(*)</b>""")

def convertstrong( element ):
	return process(element, """<strong>$(*)</strong>""")

def convertpre( element ):
	return process(element, """<pre%s>$(*)</pre>""" % (wattrs(element)))

def convertcode( element ):
	return process(element, """<code>$(*)</code>""")

def convertbreak( element ):
	return process(element, """<br />""")

def convertnewline( element ):
	return process(element, """<br />""")

def convertarrow( element ):
	arrow = element.getAttributeNS(None, "type")
	if   arrow == "left":
		return "&larr;"
	elif arrow == "right":
		return "&rarr;"
	else:
		return "&harr;"

def convertdots( element ):
	return "&hellip;"

def convertendash( element ):
	return "&ndash;"

def convertemdash( element ):
	return "&mdash;"

def convertentity( element ):
	return "&%s;" % (element.getAttributeNS( None, "num"))

# We create the processor, register the rules and define the process variable
processor      = Processor()
name2functions = {}
for symbol in filter(lambda x:x.startswith("convert"), dir()):
	name2functions[symbol] = eval(symbol)
processor.register(name2functions)
process = processor.process

# EOF



# vim: ts=4
# -----------------------------------------------------------------------------
# Project           :   Tahchee                      <http://www.ivy.fr/tachee>
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre                     <sebastien@ivy.fr>
# License           :   Revised BSD License
# -----------------------------------------------------------------------------
# Creation date     :   27-Mar-2007
# Last mod.         :   27-Mar-2007
# -----------------------------------------------------------------------------

import re, string, htmlentitydefs

NAME    = "escape"
VERSION = None
SUMMARY = """Escapes a string so that it can be safely included into an HTML \
document."""

class EscapePlugin:

	def __init__( self, site ):
		self.site = site

	def name( self ): return NAME
	def summary( self ): return SUMMARY
	def version( self ): return VERSION
	def doc( self ): return __doc__

	def install( self, localdict ):
		localdict["escape"] = self

	def toHTML( self, text ):
		"Returns the given HTML with ampersands, quotes and carets encoded"
		# NOTE: This is ripped from Django utils.html module
		if not isinstance(text, basestring):
			text = str(text)
		return text.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;').replace('"', '&quot;').replace("'", '&#39;')

	def toText( self, text ):
		"""Expands the entities found in the given text and returns it as text."""
		# NOTE: This is based on
		# <http://www.shearersoftware.com/software/developers/htmlfilter/>
		entityStart = text.find('&')
		if entityStart != -1:          # only run bulk of code if there are entities present
			preferUnicodeToISO8859 = 1 #(outputEncoding is not 'iso-8859-1')
			prevOffset = 0
			textParts = []
			while entityStart != -1:
				textParts.append(text[prevOffset:entityStart])
				entityEnd = text.find(';', entityStart+1)
				if entityEnd == -1:
					entityEnd = entityStart
					entity = '&'
				else:
					entity = text[entityStart:entityEnd+1]
					if len(entity) < 4 or entity[1] != '#':
						entity = htmlentitydefs.entitydefs.get(entity[1:-1],entity)
					if len(entity) == 1:
						if preferUnicodeToISO8859 and ord(entity) > 127 and hasattr(entity, 'decode'):
							entity = entity.decode('iso-8859-1')
					else:
						if len(entity) >= 4 and entity[1] == '#':
							if entity[2] in ('X','x'):
								entityCode = int(entity[3:-1], 16)
							else:
								entityCode = int(entity[2:-1])
							if entityCode > 255:
								entity = unichr(entityCode)
							else:
								entity = chr(entityCode)
								if preferUnicodeToISO8859 and hasattr(entity, 'decode'):
									entity = entity.decode('iso-8859-1')
					textParts.append(entity)
				prevOffset = entityEnd+1
				entityStart = text.find('&', prevOffset)
			textParts.append(text[prevOffset:])
			text = ''.join(textParts)
		return text

	def __call__( self, text ):
		return self.toHTML(text)

# EOF



#!/usr/bin/env python
# Encoding: iso-8859-1
# -----------------------------------------------------------------------------
# Project           :   Kiwi
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre                 <sebastien@type-z.org>
# -----------------------------------------------------------------------------
# Creation date     :   06-Mar-2006
# Last mod.         :   06-Mar-2006
# History           :
#                       06-Mar-2006 First implementation
#                       06-Mar-2006 First implementation
#
# Bugs              :
#                       -
# To do             :
#                       -
#

import re, textwrap, xml.dom
from templates import Processor

# Text to HTML: http://bhaak.dyndns.org/vilistextum/screenshots.html
# From <http://www.boddie.org.uk/python/HTML.html>
from xml.dom.ext.reader import HtmlLib

# We create the processor, register the rules and define the process variable
processor      = Processor()

#------------------------------------------------------------------------------
#
#  Actual element processing
#
#------------------------------------------------------------------------------

def convertHTML(element):
	return process(element, "$(*)\n")

def convertP(element):
	return "\n".join(textwrap.wrap(process(element, "$(*)"), 79)) + "\n\n"

def convertPRE(element):
	res = "\n"
	for line in process(element, "$(*)").split("\n"):
		res += ">    %s\n" % (line)
	return res + "\n"

def convertUL(element):
	res = ""
	for line in process(element, "$(*)").split("\n"):
		res += "  %s\n" % (line)
	return res + "\n\n"

def convertLI(element):
	lines = process(element, "$(*)").split("\n")
	res = " * " + lines[0]
	for line in lines[1:]:
		res += "   %s\n" % (line)
	return res + "\n"

def convertH1(element):
	res = process(element, "$(*)")
	return "\n" + res + "\n" + "=" * len(res) + "\n\n"

def convertH2(element):
	res = process(element, "$(*)")
	return "\n" + res + "\n" + "-" * len(res) + "\n\n"


name2functions = {}
for symbol in filter(lambda x:x.startswith("convert"), dir()):
	name2functions[symbol] = eval(symbol)
processor.register(name2functions)
process = processor.process

def convertDocument( text ):
	reader = HtmlLib.Reader()
	doc    = reader.fromString(text)
	return processor.generate(doc)

if __name__ == "__main__":
	import sys
	fd = file(sys.argv[1], 'rt')
	print convertDocument(fd.read())
	fd.close()

# EOF



#!/usr/bin/env python
# Encoding: iso-8859-1
# vim: tw=80 ts=4 sw=4 noet
# -----------------------------------------------------------------------------
# Project           :   Kiwi
# Module            :   Block parsers
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre                 <sebastien@type-z.org>
# License           :   Revised BSD License
# -----------------------------------------------------------------------------
# Creation date     :   19-Nov-2003
# Last mod.         :   07-Oct-2009
# -----------------------------------------------------------------------------

import re, string
from formatting import *

__doc__       = """Write module doc here"""
__pychecker__ = "unusednames=recogniseInfo,content"

EMPTY_LIST_ITEM = "Empty list item."

BLOCK_ELEMENTS = ("Block", "ListItem", "Definition", "Content", "Chapter", "Section", "Appendix")

STANDARD_LIST    = 1
DEFINITION_LIST  = 2
TODO_LIST        = 3
ORDERED_LIST     = 4

STANDARD_ITEM    = 100
TODO_ITEM        = 101
TODO_DONE_ITEM   = 102

#------------------------------------------------------------------------------
#
#  Regular expressions
#
#------------------------------------------------------------------------------

RE_BLANK          = re.compile(u"\s*", re.LOCALE|re.MULTILINE)

TITLE             = u"^\s*(==)([^=].+)$"
RE_TITLE          = re.compile(TITLE, re.LOCALE|re.MULTILINE)
TITLE_HEADER      = u"^\s*(--)([^\:]+):(.+)?$"
RE_TITLES         = re.compile(u"%s|%s" % (TITLE, TITLE_HEADER), re.LOCALE|re.MULTILINE)

SECTION_HEADING   = u"^\s*((([0-9]+|[A-z])\.)+([0-9]+|[A-z])?\.?)"
RE_SECTION_HEADING= re.compile(SECTION_HEADING, re.LOCALE)
SECTION_HEADING_ALT = u"^(\=+\s*).+$"
RE_SECTION_HEADING_ALT= re.compile(SECTION_HEADING_ALT, re.LOCALE)
SECTION_UNDERLINE = u"^\s*[\*\-\=#][\*\-\=#][\*\-\=#]+\s*$"
RE_SECTION_UNDERLINE = re.compile(SECTION_UNDERLINE, re.LOCALE|re.MULTILINE)

DEFINITION_ITEM   = u"^(\s*(\:[^\:]|[^\:])+)\:\:+\s*(\n+\s*|\s*\|\s*\n)*"
RE_DEFINITION_ITEM = re.compile(DEFINITION_ITEM, re.LOCALE|re.MULTILINE)

TAGGED_BLOCK      = u"^\s*(([^_]+\s*)(\:[^_]+)?)?(____+)\s*$"
RE_TAGGED_BLOCK   = re.compile(TAGGED_BLOCK, re.MULTILINE | re.LOCALE)
LIST_ITEM         = u"^(\s*)(-|\*\)|[0-9A-z][\)/]|\[[ \-\~xX]\])\s*"
RE_LIST_ITEM      = re.compile(LIST_ITEM, re.MULTILINE | re.LOCALE)
LIST_HEADING      = u"(^\s*[^:{().<]*:)"
RE_LIST_HEADING   = re.compile(LIST_HEADING, re.MULTILINE | re.LOCALE)
LIST_ITEM_HEADING = u"^([^:]+(:\s*\n\s*|::\s*))|([^/\\\]+[/\\\]\s*\n\s*)"
RE_LIST_ITEM_HEADING =  re.compile(LIST_ITEM_HEADING, re.MULTILINE|re.LOCALE)
RE_NUMBER          = re.compile("\d+[\)\.]")

PREFORMATTED      = u"^(\s*\>(\t|   ))(.*)$"
RE_PREFORMATTED   = re.compile(PREFORMATTED, re.LOCALE)

CUSTOM_MARKUP = u"\s*-\s*\"([^\"]+)\"\s*[=:]\s*([\w\-_]+)(\s*\(\s*(\w+)\s*\))?"
RE_CUSTOM_MARKUP = re.compile(CUSTOM_MARKUP, re.LOCALE|re.MULTILINE)

META_TYPE        = u"\s*(\w+)\s*(\((\w+)\))?"
RE_META_TYPE     = re.compile(META_TYPE, re.LOCALE|re.MULTILINE)

META_FIELD = u'(^|\n)\s*([\w\-]+)\s*:\s*'
RE_META_FIELD= re.compile(META_FIELD, re.LOCALE)
RE_META_AUTHOR_EMAIL = re.compile("\<([^>]+)\>", re.LOCALE)

REFERENCE_ENTRY    = u"\s+\[([^\]]+)]:"
RE_REFERENCE_ENTRY = re.compile(REFERENCE_ENTRY, re.LOCALE|re.MULTILINE)

TABLE_ROW_SEPARATOR    = "^\s*([\-\+]+|[\=\+]+)\s*$"
RE_TABLE_ROW_SEPARATOR = re.compile(TABLE_ROW_SEPARATOR)

LANGUAGE_CODES = ("EN", "FR", "DE", "UK" )

#------------------------------------------------------------------------------
#
#  Error messages
#
#------------------------------------------------------------------------------

ERROR_TITLE_TOO_DEEPLY_NESTED = "Title too deeply nested"

#------------------------------------------------------------------------------
#
#  BlockParser
#
#------------------------------------------------------------------------------

class BlockParser:

	def __init__( self, name ):
		self.name = name

	def recognises( self, context ):
		"""Tells wether the given block is recognised or not. This returns
		this block recognition information, or False (or None) if the block was
		not recongised."""
		return False

	def process( self, context, recogniseInfo ):
		return None

	def processText( self, context, text ):
		assert context, text
		return text

#------------------------------------------------------------------------------
#
#  ParagraphBlockParser
#
#------------------------------------------------------------------------------

class ParagraphBlockParser(BlockParser):
	"""Parses a paragraph block. This parser always recognised the given block,
	so it should not appear in the block parsers."""

	def __init__( self ):
		BlockParser.__init__(self, "Paragraph")

	def recognises( self, context ):
		return True

	def process( self, context, recogniseInfo ):
		# We make sure that the current node is a block element
		paragraph_depth = context.getBlockIndentation()
		# Here we move to the first block element that has an indentation that
		# is lower or equal to this paragraph
		while context.currentNode.nodeName not in BLOCK_ELEMENTS \
		or context.currentNode.getAttribute("_indent") \
		and int(context.currentNode.getAttribute("_indent"))>paragraph_depth:
			context.currentNode = context.currentNode.parentNode
		# If the currentNode last element is a paragraph with a higher
		# indentation than the current one, then we create a block, and set it
		# as current node (this allows to create "indented paragraphs" - the
		# equivalent of blockquotes).
		if context.currentNode.childNodes \
		and context.currentNode.childNodes[-1].nodeName == "Paragraph" \
		and context.currentNode.childNodes[-1].getAttribute("_indent") \
		and int(context.currentNode.childNodes[-1].getAttribute("_indent"))<paragraph_depth:
			block_node = context.document.createElementNS(None, "Block")
			block_node.setAttributeNS(None, "_indent", str(paragraph_depth))
			context.currentNode.appendChild(block_node)
			context.currentNode = block_node
		# Now we can process the document
		para_node = context.document.createElementNS(None, self.name)
		para_node.setAttributeNS(None, "_indent", str(paragraph_depth))
		para_node.setAttributeNS(None, "_start", str(context.blockStartOffset))
		para_node.setAttributeNS(None, "_end", str(context.blockEndOffset))
		context.parser.parseBlock(context, para_node, self.processText)
		# Now we suppress leading and trailing whitespaces
		first_text_node = para_node.childNodes[0]
		last_text_node  = para_node.childNodes[-1]
		if first_text_node.nodeType != para_node.TEXT_NODE: first_text_node = None
		if last_text_node.nodeType  != para_node.TEXT_NODE: last_text_node  = None
		# Removed first and last text nodes if empty
		if first_text_node!=None and first_text_node.data.strip()=="":
			para_node.removeChild(first_text_node)
			first_text_node = None
		if last_text_node!=None and last_text_node.data.strip()=="":
			para_node.removeChild(last_text_node)
			last_text_node = None
		# We strip the leading whitespace
		if first_text_node!=None and len(first_text_node.data)>0 and \
			first_text_node.data[0] == " ":
			first_text_node.data = first_text_node.data[1:]
		if last_text_node!=None and len(last_text_node.data)>0 and \
			last_text_node.data[-1] == " ":
			last_text_node.data = last_text_node.data[:-1]
		# FIXME: Maybe the paragraph contains text nodes with only spaces ?
		if len(para_node.childNodes)>0:
			context.currentNode.appendChild(para_node)
		else:
			context.parser.warning("Empty paragraph removed", context)

	def processText( self, context, text ):
		assert text
		text = context.parser.expandTabs(text)
		text =  context.parser.normaliseText(text)
		return text

#------------------------------------------------------------------------------
#
#  TaggedBlockParser
#
#------------------------------------------------------------------------------

class TaggedBlockParser(BlockParser):
	"""Parses a tagged block. Notes are the common example of tagged
	block."""

	def __init__( self ):
		BlockParser.__init__(self, "TaggedBlock")

	def recognises( self, context ):
		lines = filter(lambda l:l.strip(), context.currentFragment().split("\n"))
		if not lines: return
		return RE_TAGGED_BLOCK.match(lines[0])

	def _goToParent( self, thisblock, parent ):
		if not parent: return parent
		if parent.nodeName == "Block":
			return parent.parentNode
		else:
			return parent

	def process( self, context, recogniseInfo ):
		tagname  = recogniseInfo.group(2)
		tagtitle = recogniseInfo.group(3)
		# This is an opening tag
		if tagname and tagname[0] != "_":
			# TODO: Asserts we are not already in a sepcific block
			block_depth = context.getBlockIndentation()
			block_node = context.document.createElementNS(None, "Block")
			block_node.setAttributeNS(None, "type", tagname.strip().lower())
			block_node.setAttributeNS(None, "_indent",str(block_depth))
			if tagtitle:
				block_node.setAttributeNS(None, "title", tagtitle[1:].strip())
			# We get to a content node
			# Now we can process the document
			context.increaseOffset(len(recogniseInfo.group()))
			context.parser.parseBlock(context, block_node, self.processText)
			context.currentNode = self._goToParent( block_node, context.currentNode)
			context.currentNode.appendChild(block_node)
			context.currentNode = block_node
			assert context.currentNode
		# This is a closing tag
		elif tagname and tagname[0] == "_":
			while context.currentNode.nodeName != "Block":
				context.currentNode = context.currentNode.parentNode
			context.currentNode = context.currentNode.parentNode

#------------------------------------------------------------------------------
#
#  CommentBlockParser
#
#------------------------------------------------------------------------------

class CommentBlockParser(BlockParser):
	"""Parses a comment markup block."""

	def __init__( self ):
		BlockParser.__init__(self, "CommentBlock")

	def recognises( self, context ):
		assert context and context.parser.commentParser
		lines = context.currentFragment().split("\n")
		for line in lines:
			line = line.strip()
			if line and line.strip()[0]!= "#": return False
		return True

	def process( self, context, recogniseInfo ):
		context.currentNode.appendChild( context.document.createComment(
		self.processText(context, context.currentFragment())))
		context.setOffset(context.blockEndOffset)


#------------------------------------------------------------------------------
#
#  MarkupBlockParser
#
#------------------------------------------------------------------------------

class MarkupBlockParser(BlockParser):
	"""Parses a custom markup block."""

	def __init__( self ):
		BlockParser.__init__(self, "MarkupBlock")

	def recognises( self, context ):
		assert context and context.parser.markupParser
		offset, match = context.parser.markupParser.recognises(context)
		# We make sure that the recognised markup is a block markup which has
		# only whitespaces at the beginning
		if match and context.parser.markupParser.isStartTag(match) \
		and len(context.currentFragment()[:match.start()].strip())==0:
			# We parse the tag to see if it is a block tag and that it spans
			# the whole context current fragment.
			dummy_node = context.document.createElementNS(None, "Dummy")
			match_end = context.parser.markupParser.parse(context, dummy_node, match)
			# The returned matched end MUST BE GREATER than the start tag match
			# end, and there MUST BE ONLY SPACES after the match end for this
			# tag to represent a standalone block, and not a block inlined into
			# a paragraph.
			if match_end > match.end() and \
			len(context.currentFragment()[match_end:].strip())==0:
				# If there is a child node, we return it
				if len(dummy_node.childNodes)>=1:
					result_node = dummy_node.childNodes[0]
					# We take care of the attributes
					for key, value \
					in context.parseAttributes(match.group(2)).items():
						result_node.setAttributeNS(None, key, value)
					return result_node
				# Otherwise this means that the block is empty
				else: return True
			else:
				return False
		else:
			return False

	def process( self, context, recogniseInfo ):
		if recogniseInfo!=True:
			context.currentNode.appendChild(recogniseInfo)
		context.setOffset(context.blockEndOffset)


#------------------------------------------------------------------------------
#
#  TitleBlockParser
#
#------------------------------------------------------------------------------

class TitleBlockParser(BlockParser):
	"""Parses a title object"""

	def __init__( self ):
		BlockParser.__init__(self, "title")

	def recognises( self, context ):
		matches = []
		if context.content.childNodes: return None
		while not context.blockEndReached():
			match = RE_TITLES.match(context.currentFragment())
			if match!=None:
				context.increaseOffset(match.end())
				matches.append(match)
			else:
				return matches or False
		return matches

	def _processLine( self, line ):
		pass

	def process( self, context, recogniseInfo ):
		assert recogniseInfo
		for match in recogniseInfo:
			if match.group(1):
				titleNode = context.ensureElement( context.header, "Title" )
				# We get the content of the title
				titleText = Upper(match.group(2) or match.group(4))
				# We prefix with 'sub' or 'subsub' depending on the number of
				# preceding titles
				titleType  = u"sub" * len(filter(lambda n:n.nodeName.endswith("title"), titleNode.childNodes))
				titleType += u"title"
				#We add the node to the document tree
				resultNode = context.ensureElement(titleNode, titleType)
				titleNode.appendChild(resultNode)
				resultNode.appendChild(context.document.createTextNode(self.processText(context, titleText)))
			elif match.group(3):
				metaNode  = context.ensureElement( context.header, "Meta" )
				# We get the header name
				header_name = match.group(4).strip()
				header_text = match.group(5).strip()
				# We prepare the header node
				node = context.document.createElementNS(None, "meta")
				node.setAttributeNS(None, "name", header_name)
				node.appendChild(context.document.createTextNode(self.processText(context,
				header_text)))
				# And we add it to the document header
				metaNode.appendChild(node)
			else:
				raise Exception("We should not be here ! " + match.group())
		context.setOffset(context.blockEndOffset)

	def processText( self, context, text ):
		return context.parser.normaliseText(text.strip())

#------------------------------------------------------------------------------
#
#  SectionBlockParser
#
#------------------------------------------------------------------------------

class SectionBlockParser(BlockParser):
	"""Parses a section markup element."""

	def __init__( self ):
		BlockParser.__init__(self, "Section")

	def recognises( self, context ):
		# We look for the number prefix
		match     = RE_SECTION_HEADING.match(context.currentFragment())
		# We return directly if there are at least two section numbers (2.3)
		if match:
			match_underline = RE_SECTION_UNDERLINE.search(context.currentFragment())
			if match_underline: return (RE_SECTION_UNDERLINE, match_underline)
			else: return (RE_SECTION_HEADING, match) 
		# We return directly for a section prefixed by '=='
		match_alt = RE_SECTION_HEADING_ALT.match(context.currentFragment())
		if match_alt:
			return (RE_SECTION_HEADING_ALT, match_alt)
		# Or a separator followed by blank space
		match = RE_SECTION_UNDERLINE.search(context.currentFragment())
		if  match:
			# If we reached the end of the block, and that there is something
			# before, this OK
			if match.end() == context.blockEndOffset and \
			context.currentFragment()[:match.start()].strip():
				return (RE_SECTION_UNDERLINE, match)
			# Otherwise the rest must be blank
			else:
				blank_match = RE_BLANK.match(context.currentFragment()[match.end():])
				# The rest is blank, it's OK
				if blank_match.end()+match.end()+context.getOffset()\
					==context.blockEndOffset:
					return (RE_SECTION_UNDERLINE, match)
				# Otherwise there is a trailing text
				else:
					return None
		# Nothing matched
		else:
			return None

	def process( self, context, recogniseInfo ):
		context.ensureParent( ("Content", "Appendix", "Chapter", "Section") )
		matched_type, match = recogniseInfo
		section_indent = context.getBlockIndentation()
		trail = match.group().strip()
		# RULE:
		# A section underlined with '==' weights more than a section
		# underlined with '--', which weights more than a section 
		# underline with nothing. This means that if you have
		#
		#  1. One
		#  ======
		#
		#  2. Two
		#  ------
		#
		#  3. Three
		#
		# These sections will all be children of the previous section
		section_weight = trail.endswith("==") and 2 or trail.endswith("--") and 1 or 0
		#
		# FIRST STEP - We detect section text bounds
		#
		block_start  = context.blockStartOffset
		block_end    = context.blockEndOffset
		section_type = "Section"
		# We have an underlined section
		if matched_type == RE_SECTION_UNDERLINE:
			block_end = context.getOffset() + match.start()
		if matched_type == RE_SECTION_HEADING_ALT:
			block_start = context.getOffset() + match.start() + len(match.group(1))
			block_end   = context.getOffset() + match.end()
		
		# We look for a number prefix
		heading_text = context.fragment(block_start, block_end)
		prefix_match = RE_SECTION_HEADING.match(heading_text)
		dots_count   = 0
		if prefix_match:
			res         = prefix_match.group()
			dots_count  = len( filter(lambda x:x, res.split(".")) )
			block_start = context.getOffset() + prefix_match.end()
		if matched_type == RE_SECTION_HEADING_ALT:
			dots_count += len(match.group(1))
		# We make sure that we end the section before the block delimiter
		delim_match = RE_SECTION_UNDERLINE.search(context.currentFragment())
		if delim_match:
			block_end = context.getOffset() + delim_match.start()
		context.currentNode = context.getParentSection(dots_count-section_weight, section_indent)
		section_depth       = context.getDepthInSection(context.currentNode) + 1
		#
		# SECOND STEP - We create the section
		#
		section_node = context.document.createElementNS(None, section_type)
		section_node.setAttributeNS(None, "_indent", str(section_indent ))
		section_node.setAttributeNS(None, "_depth", str(section_depth))
		section_node.setAttributeNS(None, "_start", str(block_start))
		section_node.setAttributeNS(None, "_sstart", str(block_start))
		heading_node = context.document.createElementNS(None, "Heading")
		section_node.appendChild(heading_node)
		offsets = context.saveOffsets()
		context.blockEndOffset = block_end
		context.setOffset(block_start)
		context.parser.parseBlock(context, heading_node, self.processText)
		context.restoreOffsets(offsets)
		# Now we create a Content node
		content_node = context.document.createElementNS(None, "Content")
		content_node.setAttributeNS(None, "_indent", str(section_indent ))
		section_node.appendChild(content_node)
		# We append the section node and assign it as current node
		context.currentNode.appendChild(section_node)
		context.currentNode = content_node
		context.declareSection(section_node, content_node, dots_count-section_weight)

	def processText( self, context, text ):
		return context.parser.normaliseText(text.strip())

#------------------------------------------------------------------------------
#
#  DefinitionBlockParser
#
#------------------------------------------------------------------------------

class DefinitionBlockParser(BlockParser):
	"""Parses a definition markup element."""

	def __init__( self ):
		BlockParser.__init__(self, "Definition")

	def recognises( self, context ):
		return RE_DEFINITION_ITEM.match(context.currentFragment())

	def _getParentDefinition( self, node ):
		while node and node.nodeName != "Definition":
			node = node.parentNode
		return node

	def process( self, context, match ):
		parent_node = self._getParentDefinition(context.currentNode)
		_indent = context.getBlockIndentation()
		# Ensures that the parent Definition node exists
		if not parent_node:
			parent_node = context.currentNode
			while True:
				if parent_node.parentNode == None: break
				if parent_node.parentNode.nodeType == parent_node.DOCUMENT_NODE: break
				if not parent_node.getAttributeNS(None, "_indent"): break
				if int(parent_node.getAttributeNS(None, "_indent")) <= _indent: break
				parent_node = parent_node.parentNode
				if parent_node.nodeName not in BLOCK_ELEMENTS: continue
			context.currentNode = parent_node
			definition_node = context.document.createElementNS(None, "Definition")
			definition_node.setAttributeNS(None, "_indent", str(_indent))
			context.currentNode.appendChild(definition_node)
			parent_node = definition_node
		# Creates the defintion item
		definition_item = context.document.createElementNS(None, "DefinitionItem")
		definition_item.setAttributeNS(None, "_indent", str(_indent + 1))
		definition_title = context.document.createElementNS(None, "Title")
		definition_title.setAttributeNS(None, "_start", str(context.blockStartOffset))
		definition_title.setAttributeNS(None, "_end", str(context.blockStartOffset + len(match.group())))
		# Parse the content of the definition title
		offsets = context.saveOffsets()
		context.setCurrentBlock(context.blockStartOffset, context.blockStartOffset + len(match.group(1)))
		context.parser.parseBlock(context, definition_title, self.processText)
		context.restoreOffsets(offsets)
		# And continue the processing
		definition_content = context.document.createElementNS(None, "Content")
		definition_content.setAttributeNS(None, "_indent", str(_indent + 1))
		definition_content.setAttributeNS(None, "_start", str(context.blockStartOffset + match.end()))
		definition_content.setAttributeNS(None, "_end", str(context.blockEndOffset))
		definition_item.appendChild(definition_title)
		definition_item.appendChild(definition_content)
		parent_node.appendChild(definition_item)
		context.currentNode = definition_content
		# We check if there is a rest after the definition name
		rest = context.documentText[context.blockStartOffset + match.end():context.blockEndOffset]
		if not context.parser.normaliseText(rest).strip(): rest = ""
		if rest:
			offsets = context.saveOffsets()
			context.setCurrentBlock(context.blockStartOffset + match.end(), context.blockEndOffset)
			context.parser.parseBlock(context, definition_content, self.processText)
			context.restoreOffsets(offsets)

	def processText( self, context, text ):
		return context.parser.normaliseText(text)

#------------------------------------------------------------------------------
#
#  ListItemBlockParser
#
#------------------------------------------------------------------------------

class ListItemBlockParser(BlockParser):
	"""Parses a list item. A list item is an element within a list."""

	def __init__( self ):
		BlockParser.__init__(self, "ListItem")

	def recognises( self, context ):
		return RE_LIST_ITEM.match(context.currentFragment())

	def process( self, context, itemMatch ):

		context.ensureParent( ("Content", "Appendix", "Chapter", "Section", "List") )
		start_offset = context.getOffset()

		# Step 1: Determine the range of the current line item in the current
		# block. There may be more than one line item as in the following:
		# "- blah blah\n - blah blah"
		# So we have to look for another line item in the current block

		# To do so, we move the offset after the recognised list item, ie.
		# after the leading "1)", "*)", etc
		context.increaseOffset(itemMatch.end())

		# Next item match will indicate where in the current fragment the next
		# item starts.
		next_item_match = None
		if context.blockEndReached():
			context.parser.warning(EMPTY_LIST_ITEM, context)
			return

		# We search a possible next list item after the first eol
		next_eol = context.currentFragment().find("\n")
		if next_eol!=-1:
			next_item_match = RE_LIST_ITEM.search(
				context.currentFragment(), next_eol)
		else:
			next_item_match = None

		# We assign to current_item_text the text of the current item
		current_item_text = context.currentFragment()
		if next_item_match:
			current_item_text = current_item_text[:next_item_match.start()]

		# We get the list item identation
		indent = context.parser.getIndentation(
			context.parser.charactersToSpaces(itemMatch.group()))
		
		# We look for the optional list heading
		heading = RE_LIST_ITEM_HEADING.match(current_item_text)
		heading_offset = 0
		list_type   = STANDARD_LIST
		item_type   = STANDARD_ITEM
		if heading:
			# We remove the heading from the item text
			heading_offset = heading.end()
			# And update the heading variable with the heading text
			if heading.group(1):
				list_type = STANDARD_LIST
				heading_end = heading.group().rfind(":")
			else:
				list_type = DEFINITION_LIST
				heading_end = heading.group().rfind("/")

		head = itemMatch.group(2)
		if head:
			head = head.upper()
			if  head == "[ ]":
				item_type = TODO_ITEM
				list_type = TODO_LIST
			elif head == "[X]":
				item_type = TODO_DONE_ITEM
				list_type = TODO_LIST
			elif RE_NUMBER.match(head):
				list_type = ORDERED_LIST

		# The current_item_text is no longer used in the following code

		# Step 2: Now that we have the item body, and that we know if there is
		# a next item (next_item_match), we can create the list item node. To
		# do so, we first have to look for a parent "List" node in which the
		# "ListItem" node we wish to create will be inserted.

		# We want either a "List" with a LOWER OR EQUAL indent, or a "ListItem"
		# with a STRICLY LOWER indentation, or a node which is neither a List
		# or a ListItem.
		while context.currentNode.nodeName == "List" and \
		int(context.currentNode.getAttributeNS(None, "_indent"))>indent or \
		context.currentNode.nodeName == "ListItem" and \
		int(context.currentNode.getAttributeNS(None, "_indent"))>=indent:
			context.currentNode = context.currentNode.parentNode

		# If the current node is a list, then we have to create a nested list.
		# A List ALWAYS have at least one child ListItem. If the last ListItem
		# has the same indentation as our current list item, then it is a
		# sibling, otherwise it is a parent.
		if context.currentNode.nodeName == "List":
			# A List should always have a least one ListItem
			items = context._getElementsByTagName( context.currentNode, "ListItem")
			assert len(items)>0
			if int(items[-1].getAttributeNS(None, "_indent")) < indent:
				context.currentNode = items[-1]

		# We may need to create a new "List" node to hold our list items
		list_node = context.currentNode
		# If the current node is not a list, then we must create a new list
		if context.currentNode.nodeName != "List":
			list_node = context.document.createElementNS(None, "List")
			list_node.setAttributeNS(None, "_indent", str(indent))
			context.currentNode.appendChild(list_node)
			context.currentNode = list_node
		# We create the list item
		list_item_node = context.document.createElementNS(None, "ListItem")
		list_item_node.setAttributeNS(None, "_indent", str(indent))
		if item_type == TODO_ITEM:
			list_item_node.setAttributeNS(None, "todo", "true")
		elif item_type == TODO_DONE_ITEM:
			list_item_node.setAttributeNS(None, "todo", "done")
		#list_item_node.setAttributeNS(None, "_start", str(start_offset))
		if next_item_match:
			list_item_node.setAttributeNS(None, "_end", str(context.getOffset() + next_item_match.start() -1))
		else:
			list_item_node.setAttributeNS(None, "_end", str(context.blockEndOffset))
		# and the optional heading
		if heading:
			offsets = context.saveOffsets()
			heading_node = context.document.createElementNS(None, "heading")
			context.setCurrentBlock(context.getOffset(), context.getOffset()+heading_end)
			context.parser.parseBlock(context, heading_node, self.processText)
			# heading_text = context.document.createTextNode(heading)
			# heading_node.appendChild(heading_text)
			list_item_node.appendChild(heading_node)
			context.restoreOffsets(offsets)
		# and the content
		offsets = context.saveOffsets()
		if next_item_match:
			context.setCurrentBlock(heading_offset+context.getOffset() ,
				context.getOffset()+next_item_match.start())
		else:
			context.increaseOffset(heading_offset)
		# We parse the content of the list item
		old_node = context.currentNode
		# FIXME: This is necessary to assign the current node, but I do not
		# quite understand why... this needs some code review.
		context.currentNode = list_item_node
		context.parser.parseBlock(context, list_item_node, self.processText)
		context.currentNode = old_node
		context.restoreOffsets(offsets)
		# We eventually append the created list item node to the parent list
		# node
		list_node.appendChild(list_item_node)
		# We set the type attribute of the list if necesseary
		if list_type == DEFINITION_LIST:
			list_node.setAttributeNS(None, "type", "definition")
		elif list_type == TODO_LIST:
			list_node.setAttributeNS(None, "type", "todo")
		elif list_type == ORDERED_LIST:
			list_node.setAttributeNS(None, "type", "ordered")

		# And recurse with other line items
		if next_item_match:
			# We set the offset in which the next_item Match object was
			# created, because match object start and end are relative
			# to the context offset at pattern matching time.
			list_item_node = self.process(context, next_item_match)
		# Or we have reached the block end
		else:
			context.setOffset(context.blockEndOffset)

		# We set the current node to be the list item node
		context.currentNode = list_item_node
		return list_item_node

	def processText( self, context, text ):
		text = context.parser.expandTabs(text)
		text = context.parser.normaliseText(text)
		return text

#------------------------------------------------------------------------------
#
#  PreBlockParser
#
#------------------------------------------------------------------------------

class PreBlockParser( BlockParser ):
	"""Parses the content of a preformatted block, where every line is
	prefixed by '>   '."""

	def __init__( self ):
		BlockParser.__init__(self, "pre")

	def recognises( self, context ):
		for line in context.currentFragment().split("\n"):
			if line and not RE_PREFORMATTED.match(line):
				return False
		return True

	def process( self, context, recogniseInfo ):
		text = ""
		for line in context.currentFragment().split("\n"):
			match = RE_PREFORMATTED.match(line)
			if match:
				text += match.group(3) + "\n"
			else:
				text += line + "\n"
		if text[-1] == "\n": text = text[:-1]
		pre_node = context.document.createElementNS(None, self.name)
		pre_node.appendChild(context.document.createTextNode(text))
		pre_node.setAttributeNS(None, "_start", str(context.getOffset()))
		pre_node.setAttributeNS(None, "_end", str(context.blockEndOffset))
		context.currentNode.appendChild(pre_node)

class PreBlockParser2( BlockParser ):
	"""Parses the content of a preformatted block which is delimited with
	'<<<' and '>>>' characters."""

	def __init__( self ):
		BlockParser.__init__(self, "pre")

	def recognises( self, context ):
		head_lines =  context.currentFragment().split("\n")
		if not head_lines: return False
		if self.isStartLine(context, head_lines[0]):
			indent = context.parser.getIndentation(head_lines[0])
			for line in head_lines[1:]:
				if not line.replace("\t", " ").strip(): continue
				if context.parser.getIndentation(line) < indent:
					return False
		else:
			return False
		return True, indent

	def isStartLine( self, context, line ):
		line_indent = context.parser.getIndentation(line)
		if line.replace("\t", " ").strip() == "---":
			return True, line_indent
		else:
			return None

	def isEndLine( self, context, line, indent ):
		line_indent = context.parser.getIndentation(line)
		if line_indent != indent: return False
		line = line.replace("\t", " ").strip()
		return  line == "---"

	def findBlockEnd( self, context, indent ):
		# FIXME: Issue a warning if no end is found
		cur_offset = context.blockEndOffset + 1
		block_end = context.blockEndOffset
		lines = context.currentFragment().split("\n")
		if self.isEndLine(context, lines[-1], indent):
			return block_end
		while True:
			next_eol = context.documentText.find("\n", cur_offset)
			if next_eol == -1:
				break
			line = context.documentText[cur_offset:next_eol]
			if self.isEndLine(context, line, indent):
				block_end = next_eol + 1
				break
			if line.strip() and context.parser.getIndentation(line) < indent:
				break
			block_end = next_eol + 1
			cur_offset = block_end
		return block_end - 1

	def getCommonPrefix( self, linea, lineb ):
		if not lineb.replace("\t", " ").strip():
			return linea
		else:
			limit = 0
			max_limit = min(len(linea), len(lineb))
			while limit < max_limit and linea[limit] in "\t " and linea[limit] == lineb[limit]:
				limit += 1
			assert linea[:limit] == lineb[:limit]
			return linea[:limit]

	def process( self, context, recogniseInfo ):
		result = []
		indent = recogniseInfo[1]
		context.setCurrentBlockEnd(self.findBlockEnd(context, indent))
		lines = context.currentFragment().split("\n")
		lines = lines[1:-1]
		prefix   = lines[0]
		for line in lines:
			prefix = self.getCommonPrefix(prefix, line)
		for line in lines:
			line = line[len(prefix):]
			result.append(line)
		text = "\n".join(result)
		pre_node = context.document.createElementNS(None, self.name)
		pre_node.appendChild(context.document.createTextNode(text))
		pre_node.setAttributeNS(None, "_start", str(context.getOffset()))
		pre_node.setAttributeNS(None, "_end", str(context.blockEndOffset))
		context.currentNode.appendChild(pre_node)

#------------------------------------------------------------------------------
#
#  TableBlockParser
#
#------------------------------------------------------------------------------

class Table:
	"""The table class allows to easily create tables and then generate the
	XML objects from them."""
	
	def __init__( self ):
		# Table is an array of array of (char, string) where char is either
		# 'H' for header, or 'T' for text.
		self._table = []
		self._rows  = 0
		self._cols  = 0
		self._title = None
		self._id    = None
	
	def dimension( self ):
		return len(self._table[0]), len(self._table) 

	def getRow( self, y):
		return self._table[y]

	def _ensureCell( self, x, y ):
		"""Ensures that the cell at the given position exists and returns its
		pair value."""
		while y >= len(self._table): self._table.append([])
		row = self._table[y]
		while x >= len(row): row.append(["T", None])
		self._cols = max(self._cols, x+1)
		self._rows = max(self._rows, y+1)
		return row[x]
		
	def setTitle( self, title ):
		"""Sets the title for this table."""
		self._title = title.strip()

	def setID( self, id ):
		"""Sets the id for this table."""
		self._id = id.strip()

	def appendCellContent( self, x, y, text ):
		cell_type, cell_text = self._ensureCell(x,y)
		if cell_text == None:
			self._table[y][x] = [cell_type, text]
		else:
			self._table[y][x] = [cell_type, cell_text + "\n" + text]

	def headerCell( self, x, y ):
		self._table[y][x] = ["H", self._ensureCell(x,y)[1]]

	def dataCell( self, x, y ):
		self._table[y][x] = ["T", self._ensureCell(x,y)[1]]

	def isHeader( self, x, y ):
		if len(self._table) < y or len(self._table[y]) < x: return False
		row = self._table[y]
		if x>=len(row): return False
		return self._table[y][x][0] == "H"

	def getNode( self, context, processText ):
		"""Renders the table as a Kiwi XML document node."""
		table_node   = context.document.createElementNS(None, "Table")
		content_node = context.document.createElementNS(None, "Content")
		# We set the id
		if self._id:
			table_node.setAttributeNS(None, "id", self._id)
		# We take care of the title
		if self._title:
			caption_node = context.document.createElementNS(None, "Caption")
			caption_text = context.document.createTextNode(self._title)
			caption_node.appendChild(caption_text)
			table_node.appendChild(caption_node)
		# And now of the table
		for row in self._table:
			row_node = context.document.createElementNS(None, "Row")
			i = 0
			for cell_type, cell_text in row:
				is_first = i == 0
				is_last  = i == len(row) - 1
				cell_node = context.document.createElementNS(None, "Cell")
				if cell_type == "H":
					cell_node.setAttributeNS(None, "type", "header")
				# We create a temporary Content node that will stop the nodes
				# from seeking a parent content
				cell_content_node = context.document.createElementNS(None, "Content")
				if is_last and len(row) != self._cols:
					cell_node.setAttributeNS(None, "colspan", "%s" % (len(row) + 2 - i))
				new_context = context.clone()
				new_context.setDocumentText(cell_text)
				new_context.currentNode = cell_content_node
				new_context.parser.parseContext(new_context)
				# This is slightly hackish, but we simply move the nodes there
				for child in cell_content_node.childNodes:
					cell_node.appendChild(child)
				row_node.appendChild(cell_node)
				i += 1
			content_node.appendChild(row_node)
		table_node.appendChild(content_node)
		return table_node

	def __repr__( self ):
		s = ""
		i = 0
		for row in self._table:
			s += "%2d: %s\n" % (i,row)
			i += 1
		return s

class TableBlockParser( BlockParser ):
	"""Parses the content of a tables"""

	def __init__( self ):
		BlockParser.__init__(self, "table")

	def recognises( self, context ):
		lines = context.currentFragment().strip().split("\n")
		if not len(lines)>1: return False
		title_match = RE_TITLE.match(lines[0])
		if title_match:
			if not len(lines) >= 3: return False
			start_match = RE_TABLE_ROW_SEPARATOR.match(lines[1])
		else:
			start_match = RE_TABLE_ROW_SEPARATOR.match(lines[0])
		end_match = RE_TABLE_ROW_SEPARATOR.match(lines[-1])
		return start_match and end_match

	def process( self, context, recogniseInfo ):
		y = 0
		table = Table()
		# For each cell in a row
		rows = context.currentFragment().strip().split("\n")[:-1]
		# We take care of the title
		title_match = RE_TITLE.match(rows[0])
		if title_match:
			title_name = title_match.group(2).split("#",1)
			title_id   = None
			if len(title_name) == 2:
				title_name, title_id = title_name
			table.setTitle(title_name)
			table.setID(title_id)
			rows = rows[2:]
		else:
			rows = rows[1:]
		# The cells are separated by pipes (||)
		for row in rows:
			cells = []
			x = 0
			# Empty rows are simply ignored
			if not row.strip(): continue
			separator = RE_TABLE_ROW_SEPARATOR.match(row)
			# If we have not found a separator yet, we simply ensure that the
			# cell exists and appends content to it
			if not separator:
				# If the separtor is not '||' it is '|'
				if  row.find("||") == -1:
					row = row.replace("|", "||")
				for cell in row.split("||"):
					cells.append(cell)
					# We remove leading or trailing borders (|)
					if cell and cell[0]  == "|": cell = cell[1:]
					if cell and cell[-1] == "|": cell = cell[:-1]
					table.appendCellContent(x,y,cell)
					# FIXME: Weird rule
					# The default cell type is the same as the above
					# cell, if any.
					#if y>0 and table.isHeader(x,y-1):
					#	table.headerCell(x,y)
					x += 1
			# We move to the next row only when we encounter a separator. The
			# analysis of the separtor will tell you if the above cell is a
			# header or a data cell
			else:
				# FIXME: This is wrong, see below
				if separator.group(1)[0] == "=":
					row_count = table.dimension()[1]
					if row_count > 0:
						for cell in table.getRow(row_count - 1):
							cell[0] = "H"
				if separator.group(1)[0] == "-":
					row_count = table.dimension()[1]
					if row_count > 0:
						for cell in table.getRow(row_count - 1):
							cell[0] = "T"
				# FIXME: Should handle vertical tables also
				# ==================================
				# HEADER || DATA
				# =======++-------------------------
				# ....
				offset = 0
				x      = 0
				# FIXME: Here cells is always empty
				for cell in cells:
					assert None, "Should not be here"
					if separator.group(1)[offset] == "=": table.headerCell(x,y)
					else: table.dataCell(x,y)
					offset += len(cell)
					x      += 1
				y += 1
		context.currentNode.appendChild(table.getNode(context, self.processText))

#------------------------------------------------------------------------------
#
#  MetaBlockParser
#
#------------------------------------------------------------------------------

class MetaBlockParser( BlockParser ):
	"""Parses the content of a Meta block"""

	def __init__( self ):
		BlockParser.__init__(self, "Meta")
		#This is a binding from meta block section names to meta content
		#parsers
		self.field_parsers = {
			u'abstract':		self.p_abstract,
			u'acknowledgements':	self.p_ack,
			u'author':		self.p_author,
			u'authors':		self.p_author,
			u'creation':		self.p_creation,
			u'keywords':		self.p_keywords,
			u'language':		self.p_language,
			u'last-mod':		self.p_last_mod,
			u'markup':		self.p_markup,
			u'organisation':	self.p_organisation,
			u'organization':	self.p_organisation,
			u'revision':		self.p_revision,
			u'type':		self.p_type,
			u'reference':		self.p_reference
		}

	def process( self, context, recogniseInfo ):
		# Parses a particular field, with the given content
		def parse_field( field ):
			field = field.lower()
			if self.field_parsers.get(field):
				self.field_parsers.get(field)(context, context.currentFragment())
			else:
				context.parser.warning("Unknown Meta field: " + last_field,
				context)

		match  = True
		offset = 0
		last_field = None
		# Iterates through the fields
		while match != None:
			match = RE_META_FIELD.search(context.currentFragment(), offset)
			if match:
				if last_field != None:
					offsets = context.saveOffsets()
					# We set the current fragment to be the field value
					context.setCurrentBlock( context.getOffset() + offset,
					context.getOffset() + match.start() )
					parse_field(last_field)
					context.restoreOffsets(offsets)
				last_field = match.group(2)
				offset = match.end()

		# And parse the last field
		if last_field != None:
			offsets = context.saveOffsets()
			context.setCurrentBlock( context.getOffset() + offset,
			context.blockEndOffset )
			parse_field(last_field)
			context.restoreOffsets(offsets)
		else:
			context.parser.warning("Empty Meta block.", context)

	# Field parsers __________________________________________________________

	def p_abstract( self, context, content ):
		old_node = context.currentNode 
		abstract_node = context.document.createElementNS(None, "Abstract")
		context.currentNode = abstract_node
		context.parser.parseBlock(context, abstract_node, self.processText)
		context.currentNode  = old_node
		context.header.appendChild(abstract_node)

	def p_ack( self, context, content ):
		old_node = context.currentNode 
		ack_node = context.document.createElementNS(None, "Acknowledgement")
		context.currentNode = ack_node
		context.parser.parseBlock(context, ack_node, self.processText)
		context.currentNode  = old_node
		context.header.appendChild(ack_node)

	def p_author( self, context, content ):
		authors_node = context.document.createElementNS(None, "Authors")
		text = self._flatify(content).strip()
		# Cuts the trailing dot if present
		if text[-1]==u'.': text=text[:-1]
		for author in text.split(','):
			author_node = context.document.createElementNS(None, "person")
			# We take care of email
			email_match = RE_META_AUTHOR_EMAIL.search(author)
			if email_match:
				author = author[:email_match.start()]
				author_node.setAttributeNS(None, "email", email_match.group(1))
			text_node   = context.document.createTextNode(author.strip())
			author_node.appendChild(text_node)
			authors_node.appendChild(author_node)
		context.header.appendChild(authors_node)
	
	def p_creation( self, context, content ):
		creation_node = context.document.createElementNS(None, "creation")
		if self._parseDateToNode( context, content, creation_node ):
			context.header.appendChild(creation_node)
	
	def _parseDateToNode( self, context, content, node ):
		content = content.strip()
		date = content.split("-")
		for elem in date:
			format = None
			try:
				format = "%0" + str(len(elem)) + "d"
				format = format % (int(elem))
			except:
				pass
			if len(date)!=3 or format != elem:
				context.parser.error("Malformed date meta field: " + content,
				context)
				context.parser.tip("Should be YYYY-MM-DD", context)
				return False
		date = map(lambda x:int(x), date)
		if date[1] < 1 or date[1] > 12:
			context.parser.warning("Bad month number: " + str(date[1]),
			context)
		if date[2] < 1 or date[2] > 31:
			context.parser.warning("Bad day number: " + str(date[2]),
			context)
		node.setAttributeNS(None, "year",  str(date[0]))
		node.setAttributeNS(None, "month", str(date[1]))
		node.setAttributeNS(None, "day",   str(date[2]))
		return True

	def p_keywords( self, context, content ):
		keywords_node = context.document.createElementNS(None, "Keywords")
		text = self._flatify(content).strip()
		# Cuts the trailing dot if present
		if text[-1]==u'.': text=text[:-1]
		for keyword in text.split(','):
			keyword_node = context.document.createElementNS(None, "keyword")
			text_node   = context.document.createTextNode(keyword.strip())
			keyword_node.appendChild(text_node)
			keywords_node.appendChild(keyword_node)
		context.header.appendChild(keywords_node)

	def p_last_mod( self, context, content ):
		lastmod_node = context.document.createElementNS(None, "modification")
		if self._parseDateToNode( context, content, lastmod_node ):
			context.header.appendChild(lastmod_node)

	def p_revision( self, context, content ):
		revision_node = context.document.createElementNS(None, "revision")
		text_node   = context.document.createTextNode(content.strip())
		revision_node.appendChild(text_node)
		context.header.appendChild(revision_node)

	def p_type( self, context, content ):
		match = RE_META_TYPE.match(content)
		if match:
			style_node = context.document.createElementNS(None, "type")
			style_node.setAttributeNS(None, "name", match.group(1).lower())
			if match.group(3):
				style_node.setAttributeNS(None, "style", match.group(3).lower())
			context.header.appendChild(style_node)
		else:
			context.parser.warning("Malformed meta type field: " + content,
			context)

	def p_reference( self, context, content ):
		ref_node = context.document.createElementNS(None, "reference")
		ref_node.setAttributeNS(None, "id", content)
		context.header.appendChild(ref_node)

	def p_language( self, context, content ):
		lang = content.strip()[0:2].upper()
		lang_node = context.document.createElementNS(None, "language")
		#We assign the language code
		if len(lang)>=2 and lang.upper()[0:2] in LANGUAGE_CODES:
			lang_code = unicode(lang.upper()[0:2])
		else:
			lang_code = "UK"
		lang_node.setAttributeNS(None, "code", lang_code)
		context.header.appendChild(lang_node)

	def p_organisation( self, context, content ):
		old_node = context.currentNode 
		org_node = context.document.createElementNS(None, "Organisation")
		context.currentNode = org_node
		context.parser.parseBlock(context, org_node, self.processText)
		context.currentNode  = old_node
		context.header.appendChild(org_node)

	def p_markup( self, context, content ):
		"""Parses custom markup and registers the new parsers in the current
		Kiwi parser"""
		# TODO
		match = 1
		start = 0
		end   = len(content)
		custom_markup = RE_CUSTOM_MARKUP
		while match!=None and start<end:
			match = custom_markup.search(content,start)
			if match:
				regexp  = match.group(1)
				element = match.group(2)
				option  = match.group(4)
				if option == None:
					self.parser.txt_parsers.append(InlineParser(self.parser,
					element, regexp))
				elif option.lower() == u"empty":
					self.parser.txt_parsers.append(EmptyInlineParser(self.parser,
					element, regexp))
				else:
					#FIXME: OUTPUT ERROR FOR UNKNOWN OPTION
					pass
				start = match.end()

	def _flatify( self, text ):
		new_text = u""
		for line in text.split(): new_text += line+u" "
		return new_text

	def processText( self, context, text ):
		assert text
		text = context.parser.expandTabs(text)
		text =  context.parser.normaliseText(text)
		return text

#------------------------------------------------------------------------------
#
# ReferenceEntryBlockParser
#
#------------------------------------------------------------------------------

class ReferenceEntryBlockParser( BlockParser ):
	"""Parses the content of a Reference entry"""

	def __init__( self ):
		BlockParser.__init__(self, "Entry")

	def recognises( self, context ):
		assert context
		return RE_REFERENCE_ENTRY.match(context.currentFragment())

	def process( self, context, match ):
		offsets = context.saveOffsets()
		ranges  = []
		offset  = 0
		# We get the start and end offsets of entry blocks
		while True:
			m = RE_REFERENCE_ENTRY.search(context.currentFragment(), offset)
			if not m: break
			ranges.append((m, m.start()))
			offset = m.end()
		ranges.append((None, len(context.currentFragment())))
		new_ranges = []
		for i in range(0, len(ranges)-1):
			new_ranges.append((ranges[i][0], ranges[i][1], ranges[i+1][1]))
		ranges = new_ranges
		# We loop for each found reference entry
		for match, start_offset, end_offset in ranges:
			entry_name  = match.group(1)
			# We set the current block and process it
			sub_offsets = context.saveOffsets()
			context.setCurrentBlock(context.getOffset() + match.end(), context.getOffset() + end_offset)
			entry_node = context.document.createElementNS(None, "Entry")
			entry_node.setAttributeNS(None, "id", entry_name)
			context.parser.parseBlock(context, entry_node, self.processText)
			context.references.appendChild(entry_node)
			context.restoreOffsets(sub_offsets)
		context.restoreOffsets(offsets)

# EOF



#!/usr/bin/env python
# Encoding: iso-8859-1
# vim: tw=80 ts=4 sw=4 noet
# -----------------------------------------------------------------------------
# Project           :   Kiwi
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre                 <sebastien@type-z.org>
# -----------------------------------------------------------------------------
# Creation date     :   19-Nov-2007
# Last mod.         :   19-Nov-2007
# -----------------------------------------------------------------------------

import re, xml.dom
import sys
from formatting import *
import templates

#------------------------------------------------------------------------------
#
#  Actual element processing
#
#------------------------------------------------------------------------------

def convertDocument(element):
	return process(element, """\
$(Header:title)
$(Content)
$(References)
""")

def convertHeader( element ):
	return process(element, "---+ $(Title/title)\n---++ $(Title/subtitle)\n" )

def convertHeading( element ):
	return process(element, "$(*)")

def convertSection( element ):
	level = int(element.getAttributeNS(None, "_depth"))
	prefix = "---+" + "+" * level
	return process(element,
		prefix + " $(Heading)\n\n"
		+ "$(Content:section)"
	)

def convertReferences( element ):
	return process(element, """  $(Entry)""")

def convertParagraph( element ):
	return process(element, """$(*)\n\n""")

def convertParagraph_cell( element ):
	return process(element, """$(*)\n""")

def convertList( element ):
	is_todo = element.getAttributeNS(None, "type")
	attrs = [""]
	return process(element, """$(*)\n\n""")

def convertListItem( element ):
	attrs   = [""]
	is_todo = element.getAttributeNS(None, "todo")
	return process(element, """   * $(*)\n""")

def convertTable( element ):
	return process(element, """$(Content:table)\n\n""")

def convertDefinition( element ):
	return process(element, """$(*)\n\n""")

def convertDefinitionItem( element ):
	return process(element, """   $ $(Title) : $(Content)\n""")

def convertRow( element ):
	try: index = element.parentNode.childNodes.index(element) % 2 + 1
	except: index = 0 
	classes = ( "", "even", "odd" )
	return process(element, """$(*) |\n""")

def convertCell( element ):
	suffix     = ""
	if element.hasAttributeNS(None, "colspan"):
		suffix = " " + "|" * (int(element.getAttributeNS(None,"colspan")) - 2)
	cell = process(element, "$(*:cell)")[:-1]
	return ("| " + cell + suffix)

def convertBlock( element ):
	title = element.getAttributeNS(None,"title") or element.getAttributeNS(None, "type") or ""
	css_class = ""
	if title:
		css_class=" class='ann%s'" % (element.getAttributeNS(None, "type").capitalize())
		title = "<div class='title'>%s</div>"  % (title)
		div_type = "div"
	elif not element.getAttributeNS(None, "type"):
		div_type = "blockquote"
	return process(element, """<%s%s>%s<div class='content'%s>$(*)</div></%s>""" % (div_type, css_class, title, "", div_type))

def stringToTarget( text ):
	return text.replace("  ", " ").strip().replace(" ", "-").upper()

def convertlink( element ):
	if element.getAttributeNS(None, "type") == "ref":
		return process(element, """[[#%s][$(*)]]""" %
		(stringToTarget(element.getAttributeNS(None, "target"))))
	else:
		# TODO: Support title
		return process(element, """[[%s][$(*)]]""" %
		(element.getAttributeNS(None, "target")))

def converttarget( element ):
	name = element.getAttributeNS(None, "name")
	return process(element, """#%s $(*)""" % (stringToTarget(name)))


def convertemail( element ):
	mail = ""
	for c in  process(element, """$(*)"""):
		mail += c
	return """[[mailto:%s][%s]]""" % (mail, mail)

def converturl( element ):
	return process(element, """[[$(*)][(*)]]""")

def convertterm( element ):
	return process(element, """*$(*)*""")

def convertquote( element ):
	return process(element, """''_$(*)_''""")

def convertstrong( element ):
	return process(element, """__$(*)__""")

def convertpre( element ):
	return process(element, """<verbatim>\n$(*)\n</verbatim>\n\n""")
	
def convertcode( element ):
	return process(element, """=$(*)=""")

def convertemphasis( element ):
	return process(element, """_$(*)_""")

def convertbreak( element ):
	return process(element, """ """)

def convertnewline( element ):
	return process(element, """ """)

def convertarrow( element ):
	arrow = element.getAttributeNS(None, "type")
	if   arrow == "left":
		return "<--"
	elif arrow == "right":
		return "-->"
	else:
		return "<-->"

def convertdots( element ):
	return "..."

def convertendash( element ):
	return " -- "

def convertemdash( element ):
	return " -- "

def convertentity( element ):
	return "&%s;" % (element.getAttributeNS( None, "num"))

# We create the processor, register the rules and define the process variable
processor      = templates.Processor()
name2functions = {}
for symbol in filter(lambda x:x.startswith("convert"), dir()):
	name2functions[symbol] = eval(symbol)
processor.register(name2functions)
process = processor.process

# EOF



# vim: ts=4
# -----------------------------------------------------------------------------
# Project           :   Tahchee                      <http://www.ivy.fr/tachee>
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre                     <sebastien@ivy.fr>
# License           :   Revised BSD License
# -----------------------------------------------------------------------------
# Creation date     :   23-Mar-2006
# Last mod.         :   26-Jul-2008
# -----------------------------------------------------------------------------

import os, sys, StringIO
try:
	import kiwi.main as kiwi
except ImportError:
	import _kiwi.main as kiwi

NAME    = "kiwi"
VERSION = None
SUMMARY = "Kiwi markup to HTML conversion functions."

class KiwiPlugin:

	def __init__( self, site ):
		self.site = site

	def name( self ): return NAME
	def summary( self ): return SUMMARY
	def version( self ): return VERSION
	def doc( self ): return __doc__

	def install( self, localdict ):
		localdict["kiwi"] = self
	
	def include( self, path ):
		if not path[0] == "/":
			path = self.site.pagesDir + "/" + path
		if not os.path.exists(path):
			self.site.error("Included file not found: "+ path)
		if kiwi:
			_, r = kiwi.run("-m --body-only " + path, noOutput=True)
			return r
		else:
			self.site.warn("Kiwi is not available, but you used the $site.kiwi function")
			self.site.info("You can get Kiwi from <http://www.ivy.fr/kiwi>")
			return path

	def process( self, text, level=0 ):
		"""If Kiwi is available, the given text will be interpreted as Kiwi
		markup and HTML will be generated from it. If Kiwi is not available, a
		warning will be issued, and the text will be displayed as-is."""
		if kiwi:
			try:
				s = StringIO.StringIO(text)
				_, r = kiwi.run("-m --body-only --level=%s --" % (level), s, noOutput=True)
				s.close()
				return r
			except:
				self.site.error("Can't process kiwi markup in file %s" % self.site.page)
				raise
		else:
			self.site.warn("Kiwi is not available, but you used the $site.kiwi function")
			info("You can get Kiwi from <http://www.ivy.fr/kiwi>")
			return text

	def __call__( self, text, level=0 ):
		return self.process(text,level)

# EOF



# vim: ts=4
# -----------------------------------------------------------------------------
# Project           :   Tahchee                      <http://www.ivy.fr/tachee>
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre                     <sebastien@ivy.fr>
# License           :   Revised BSD License
# -----------------------------------------------------------------------------
# Creation date     :   14-Mar-2008
# Last mod.         :   14-Mar-2008
# -----------------------------------------------------------------------------

import os, sys, StringIO
try:
	import pamela.engine as pamela
except ImportError:
	import _pamela.engine as pamela

NAME    = "pamela"
VERSION = None
SUMMARY = "Pamela markup to HTML conversion functions."

class PamelaPlugin:

	def __init__( self, site ):
		self.site = site

	def name( self ): return NAME
	def summary( self ): return SUMMARY
	def version( self ): return VERSION
	def doc( self ): return __doc__

	def install( self, localdict ):
		localdict["pamela"] = self
	
	def include( self, path ):
		if not path[0] == "/":
			path = self.site.pagesDir + "/" + path
		if pamela:
			r = pamela.run(path)
			return r
		else:
			self.site.warn("Pamela is not available, but you used the $site.pamela function")
			self.site.info("You can get Kiwi from <http://www.ivy.fr/pamela>")
			return path

	def process( self, text ):

		if pamela:
			try:
				return pamela.Parser().parseText(text)
			except Exception, e:
				self.site.error("Can't process pamela markup " + str(e))
				raise
		else:
			self.site.warn("Pamela is not available, but you used the $site.pamela function")
			info("You can get Pamela from <http://www.ivy.fr/pamela>")
			return text

	def __call__( self, text):
		return self.process(text)

# EOF



#!/usr/bin/env python
# Encoding: iso-8859-1
# vim: tw=80 ts=4 sw=4 noet
# -----------------------------------------------------------------------------
# Project           :   Kiwi
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre                 <sebastien@type-z.org>
# License           :   Revised BSD License
# -----------------------------------------------------------------------------
# Creation date     :   07-Fev-2006
# Last mod.         :   05-Aug-2008
# -----------------------------------------------------------------------------

import os, sys

import re, string, operator, getopt, codecs

# We use 4Suite domlette
#import Ft.Xml.Domlette
#dom = Ft.Xml.Domlette.implementation
# We use minidom implementation
import xml.dom.minidom
dom = xml.dom.minidom.getDOMImplementation()

from inlines import *
from blocks  import *

#------------------------------------------------------------------------------
#
#  Globals
#
#------------------------------------------------------------------------------

# How many spaces a tab represent.
TAB_SIZE = 4

#------------------------------------------------------------------------------
#
#  Regular expressions
#
#------------------------------------------------------------------------------

RE_BLOCK_SEPARATOR = re.compile(u"[ \t\r]*\n[ \t\r]*\n", re.MULTILINE | re.LOCALE)
RE_SPACES = re.compile(u"[\s\n]+", re.LOCALE|re.MULTILINE)
RE_TABS = re.compile("\t+")
ATTRIBUTE = u"""(\w+)\s*=\s*('[^']*'|"[^"]*")"""
RE_ATTRIBUTE = re.compile(ATTRIBUTE, re.LOCALE|re.MULTILINE)

#------------------------------------------------------------------------------
#
#  Parsing context
#
#------------------------------------------------------------------------------

class Context:
	"""The context stores information on the currently processed document. It
	has the following attributes:

		- document: a reference to the current XML document.
		- rootNode: the current XML document root  node.
		- header: the XML node corresponding to the header.
		- content: the XML node corresponding to the content.
		- references: the XML node corresponding to the references.
		- appendices: the XML node corresponding to the appendices.
		- currentNode: the XML node to which attributes/nodes are added during
	  	parsing.
		- blockStartOffset: the offset in the text where the currently parsed block
	  	starts.
		- blockEndOffset: the offset in the text where the currently parsed block
	  	ends.
		- parser: a reference to the Kiwi parser instance using the context.
	"""

	def __init__( self, documentText, markOffsets=False ):
		self.document = None
		self.rootNode = None
		self.header   = None
		self.content  = None
		self.references = None
		self.appendices = None
		self.currentNode = None
		self._offset = 0
		self.blockStartOffset = 0
		self.blockEndOffset = -1
		self.setDocumentText(documentText)
		self._currentFragment = None
		self.parser = None
		self.markOffsets = markOffsets
		self.sections = []
		# These are convenience attributes used to make it easy for
		# post-verification of the links (are they all resolved)
		self._links   = []
		self._targets = []

	def _getElementsByTagName(self, node, name):
		if node.nodeType == node.ELEMENT_NODE and \
		   node.localName == name:
			result = [node]
		else:
			result = []
		for child in node.childNodes:
			result.extend(self._getElementsByTagName(child, name))
		return result

	def ensureElement( self, node, elementName, index=0 ):
		"""Ensures that the given element exists in the given node at the given
		index."""
		result = self._getElementsByTagName(node, elementName)
		if len(result)<=index:
			newElement = self.document.createElementNS(None, elementName)
			node.appendChild(newElement)
			return newElement
		else:
			return result[index]

	def ensureParent( self, parentNames, predicate=lambda x:True ):
		"""Ensures that the parent node name is one of the following name. This
		is useful for block parsers which want to ensure that their parent is a
		specific node"""
		if self.currentNode!=None:
			while ( self.currentNode.nodeName not in parentNames 
				or not predicate(self.currentNode)
				) and self.currentNode.nodeName!="Document":
				if self.currentNode.parentNode:
					self.currentNode = self.currentNode.parentNode
				else:
					return

	def declareSection( self, node, contentNode, depth ):
		"""Declares a section node with the given depth (which can be
		negative)."""
		self.sections.append((node, contentNode, depth))

	def getParentSection( self, depth, indent ):
		"""Gets the section that would be the parent section for the
		given depth."""
		for i in range(len(self.sections)-1,-1,-1):
			section = self.sections[i]
			section_node = section[0]
			section_content = section[1]
			section_depth = section[2]
			section_indent = int(section_node.getAttributeNS(None, "_indent"))
			if indent > section_indent:
				return section_content 
			elif section_indent <= indent and section_depth < depth:
				return section_content
		return self.content

	def getDepthInSection( self, node ):
		"""Returns the number of parent sections of the given node."""
		sections = 0
		while node.parentNode:
			if node.nodeName == "Section":
				sections += 1
			node = node.parentNode
		return sections

	def setDocumentText( self, text ):
		"""Sets the text of the current document. This should only be called
		at context initialisation."""
		if not type(text) == type(u""):
			text = unicode(text)
		self.documentText = text
		self.documentTextLength = len(text)
		self.blockEndOffset = self.documentTextLength
		self.setOffset(0)

	def setOffset( self, offset ):
		"""Sets the current offset."""
		self._offset = offset
		self._currentFragment = None

	def getOffset(self):
		"""Returns the current offset."""
		return self._offset

	def increaseOffset( self, increase ):
		"""Increases the current offset"""
		# We get the current fragment, because it will be freed by changing the
		# offset
		fragment = self.currentFragment()
		self.setOffset(self.getOffset()+increase)
		# We optimise current fragment access, by restoring it with a proper
		# value when possible
		if self.getOffset()<self.blockEndOffset:
			self._currentFragment = fragment[increase:]

	def decreaseOffset( self, decrease ):
		"""Decreases the offset."""
		self.increaseOffset(-decrease)

	def fragment( self, start, end ):
		"""Returns the text fragment that starts and ends at the given
		offsets."""
		return self.documentText[start:end]

	def currentFragment( self ):
		"""Returns the current text fragment, from the current offset to the
		block end offset."""
		assert self.getOffset()<self.blockEndOffset,\
		"Offset greater than block end: %s >= %s" % (self.getOffset(), self.blockEndOffset)
		if not self._currentFragment:
			self._currentFragment = \
			self.documentText[self.getOffset():self.blockEndOffset]
		return self._currentFragment

	def documentEndReached( self ):
		"""Returns true if the current offset is greater than the document
		length"""
		return self.getOffset() >= self.documentTextLength

	def blockEndReached( self ):
		"""Returns true when the current offset has exceeded the current block
		end offset"""
		return self.getOffset() >= self.blockEndOffset
	
	def offsetInBlock( self, offset ):
		"""Tells if the givne offset is in the current block."""
		return self.blockStartOffset <= offset <= self.blockEndOffset

	def setCurrentBlock( self, startOffset, endOffset ):
		"""Sets the start and end offset of the current block. The current
		offset is set to the current block start."""
		if endOffset <= 0: endOffset += self.documentTextLength
		assert startOffset>=0
		assert endOffset<=self.documentTextLength
		assert startOffset<=endOffset, "Start offset too big: %s > %s" % (startOffset, endOffset)
		self.setOffset(startOffset)
		self.blockStartOffset = startOffset
		self.blockEndOffset = endOffset
		self._currentFragment = None

	def setCurrentBlockEnd( self, endOffset ):
		assert endOffset >= self.blockStartOffset
		self.blockEndOffset = endOffset
		self._currentFragment = None

	def getBlockIndentation( self ):
		"""Returns the indentation of the current block."""
		return self.parser.getIndentation(
			self.documentText[self.blockStartOffset:self.blockEndOffset])

	def saveOffsets( self ):
		"""Returns a value that can be later used with the restoreOffsets
		method to restore the offsets as they were."""
		return (self.blockStartOffset, self.getOffset(), self.blockEndOffset)

	def restoreOffsets( self, offsets ):
		"""Takes a value returned by saveOffsets and restores the offsets as
		they were."""
		self.blockStartOffset = offsets[0]
		self.setOffset( offsets[1] )
		self.blockEndOffset = offsets[2]

	def clone( self ):
		"""Returns a clone of the current context, which can be changed safely
		without modifying the current context."""
		clone = Context(self.documentText)
		clone.document    = self.document
		clone.rootNode    = self.rootNode
		clone.header      = self.header
		clone.content     = self.content
		clone.references  = self.references
		clone.appendices  = self.appendices
		clone.currentNode = self.currentNode
		clone.parser      = self.parser
		clone.document    = self.document
		clone.setOffset(self.getOffset())
		clone.setCurrentBlock(self.blockStartOffset, self.blockEndOffset)
		return clone

	def findNextInline( self, inlineParsers ):
		"""Finds the next inline in the given context, using the given list of
		inline parsers. This does not modifies the context.

		Returns either None or a triple (offset, information, parser), where
		the offset is relative to the context offset and indicates the start
		offset where the parser recognised its tag and information is the
		information returned by the parser."""
		# We look for the inline parser that parses an inline with the lowest
		# offset
		results = []
		for inlineParser in inlineParsers:
			match_offset, result = inlineParser.recognises(self)
			if match_offset!=None:
				assert match_offset >= 0
				results.append((match_offset, result, inlineParser))
		matchedResult = None
		minimumOffset = self.documentTextLength+1
		# We get the closest matching parser
		for result in results:
			if result[0]<minimumOffset:
				minimumOffset = result[0]
				matchedResult = result
		return matchedResult
		
	def parseAttributes( self, text ):
		"""Parses attributes expressed in the given text. Attributes have the
		following form: ATTRIBUTE="VALUE" and are separated by spaces."""
		if not text: return {}
		text = text.strip()
		attributes = {}
		match  = True
		# We parse attributes
		while match and text:
			match = RE_ATTRIBUTE.match(text)
			if not match: break
			attributes[match.group(1)] = match.group(2)[1:-1]
			offset = match.end()
			text = text[match.end():].strip()
		return attributes

#------------------------------------------------------------------------------
#
#  Kiwi parser
#
#------------------------------------------------------------------------------

class Parser:

	def __init__( self, baseDirectory, inputEncoding="utf8", outputEncoding="utf8" ):
		self.blockParsers  = []
		self.inlineParsers = []
		self.customParsers = {}
		self.baseDirectory = baseDirectory
		self.inputEncoding = inputEncoding
		self.outputEncoding = outputEncoding
		self.createBlockParsers()
		self.createInlineParsers()
		self.createCustomParsers()

	def createBlockParsers( self ):
		self.blockParsers.extend((
			CommentBlockParser(),
			MarkupBlockParser(),
			PreBlockParser(),
			PreBlockParser2(),
			TableBlockParser(),
			ReferenceEntryBlockParser(),
			TitleBlockParser(),
			SectionBlockParser(),
			DefinitionBlockParser(),
			ListItemBlockParser(),
			ReferenceEntryBlockParser(),
			TaggedBlockParser(),
		))
		self.defaultBlockParser = ParagraphBlockParser()
	
	def createCustomParsers( self ):
		#self.customParsers["Meta"] = MetaBlockParser()
		self.customParsers["pre"]  = PreBlockParser()
		#self.customParsers["table"]= TableBlockParser()
		pass

	def createInlineParsers( self ):
		# Escaped and markup inline parser are the most important parsers,
		# because they MUST be invoked before any other.
		self.escapedParser = EscapedInlineParser()
		self.commentParser = CommentInlineParser()
		self.markupParser  = MarkupInlineParser()
		def normal( x,y ): return self.normaliseText(x.group(1))
		def term  ( x,y ): return self.normaliseText(x.group()[1:-1])
		self.inlineParsers.extend((
			self.escapedParser,
			self.commentParser,
			self.markupParser,
			EscapedStringInlineParser(),
			InlineParser("email",		RE_EMAIL),
			InlineParser("url",			RE_URL),
			InlineParser("url",			RE_URL_2),
			EntityInlineParser(),
			LinkInlineParser(),
			PreInlineParser(),
			TargetInlineParser(),
			InlineParser("code",		RE_CODE_2),
			InlineParser("code",		RE_CODE),
			InlineParser("term",		RE_TERM,     normal),
			InlineParser("strong",		RE_STRONG,   normal),
			InlineParser("emphasis",	RE_EMPHASIS, normal),
			InlineParser("quote",		RE_QUOTED,   normal),
			InlineParser("code",		RE_CODE_3, requiresLeadingSpace=True),
			InlineParser("citation",	RE_CITATION, normal),
			# Special characters
			InlineParser("break",		RE_BREAK),
			InlineParser(None,			RE_SWALLOW_BREAK),
			InlineParser("newline",		RE_NEWLINE),
			InlineParser("dots",		RE_DOTS),
			ArrowInlineParser(),
			InlineParser("endash",		RE_LONGDASH),
			InlineParser("emdash",		RE_LONGLONGDASH),
		))

	def _initialiseContextDocument(self, context):
		"""Creates the XML document that will be populated by Kiwi
		parsing."""
		document  = dom.createDocument(None,None,None)
		root_node = document.createElementNS(None, "Document")
		document.appendChild(root_node)
		context.rootNode = root_node
		context.document = document
		context.header   = document.createElementNS(None, "Header")
		context.content  = document.createElementNS(None, "Content")
		context.references = document.createElementNS(None, "References")
		context.appendices = document.createElementNS(None, "Appendices")
		root_node.appendChild(context.header)
		root_node.appendChild(context.content)
		root_node.appendChild(context.references)
		root_node.appendChild(context.appendices)
		context.currentNode = context.content

	# EXCEPTIONS_______________________________________________________________

	def _print( self, message, context ):
		text   = context.documentText[:context.getOffset()]
		line   = len(text.split("\n"))
		offset = context.getOffset() - text.rfind("\n") - 1
		message = unicode(message % (line, offset) + "\n")
		sys.stderr.write(message.encode("iso-8859-1"))

	def warning( self, message, context ):
		self._print( "WARNING at line %4d, character %3d: "+message, context)

	def tip( self, message, context ):
		self._print( "%4d:%3d >> " +message, context)

	def error( self, message, context ):
		self._print( "ERROR at line %4d, character %3d: "+message, context)

	# PARSING__________________________________________________________________

	def parse( self, text, offsets=False ):
		"""Parses the given text, and returns an XML document. If `offsets` is
		set to True, then all nodes of the document are annotated with their
		position in the original text as well with a number. The document will
		also have an `offsets` attribute that will contain a list of (start,
		end) offset tuples for each element."""
		# Text MUST be unicode
		assert type(text) == type(u"")
		context = Context(text, markOffsets=offsets)
		self._initialiseContextDocument(context)
		context.parser = self
		while not context.documentEndReached():
			self._parseNextBlock(context)
		# We remove unnecessary nodes
		for node in ( context.header, context.content, context.references,
		context.appendices ):
			if len(node.childNodes) == 0:
				context.rootNode.removeChild(node)
		if offsets:
			context.offsets = self._updateElementOffsets(context, offsets=[])
		return context.document

	def parseContext( self, context ):
		while not context.documentEndReached():
			self._parseNextBlock(context)

	def _parseNextBlock( self, context, end=None ):
		"""Parses the block identified in the given context, ending at the given
		'end' (if 'end' is not None)."""
		assert context!=None
		# This variable indicates if at least one block parser recognised the
		# current block
		recognised = None
		# We find block start and end
		block_start_offset = context.getOffset()
		block_end_offset, next_block_start_offset = \
			self._findNextBlockSeparator(context)
		# If we specify the end
		if end != None:
			block_end_offset = min(end, block_end_offset)
		# If the block is an empty block (a SEPARATOR), we try to find the
		# parent node
		if block_end_offset == block_start_offset:
			# We rewind until we find a "Content" block
			while context.currentNode.nodeName != "Content" and \
			context.currentNode.parentNode != None:
				context.currentNode = context.currentNode.parentNode
		# Otherwise we set the current block and process it
		else:
			context.setCurrentBlock(block_start_offset, block_end_offset)
			assert block_start_offset < block_end_offset <= next_block_start_offset
			# We first look for a block parser that recognises the current
			# context
			assert len(self.blockParsers)>0
			for blockParser in self.blockParsers:
				context.setOffset(block_start_offset)
				recognised = blockParser.recognises(context)
				context.setOffset(block_start_offset)
				if recognised: break
			# If no block parser was recognised, we used the default block
			# parser
			if not recognised:
				blockParser = self.defaultBlockParser
				recognised = self.defaultBlockParser.recognises(context)
				context.setOffset(block_start_offset)
				assert recognised
			start_offset = str(context.getOffset())
			blockParser.process(context, recognised)
			# Just in case the parser modified the end offset, we update
			# the next block start offset
			next_block_start_offset  = context.blockEndOffset
			node = context.currentNode
		# Anyway, we set the offset to the next block start
		context.setOffset(next_block_start_offset)

	def parseBlock( self, context, node, textProcessor ):
		"""Parses the current block, looking for the inlines it may contain."""
		#if context.markOffsets and not node.getAttributeNS(None,"_start"):
		#	node.setAttributeNS(None, "_start", str(context.getOffset()))
		while not context.blockEndReached():
			self._parseNextInline(context, node, textProcessor)
		#if context.markOffsets and not node.getAttributeNS(None,"_end"):
		#	node.setAttributeNS(None, "_end", str(context.getOffset()))

	def _parseNextInline( self, context, node, textProcessor ):
		"""Parses the content of the current block, starting at the context
		offset, modifying the given node and updating the context offset.
		This returns the a triple (offset, information, parser) where
		information is the result of the parser `recognises' method."""
		assert context and node and textProcessor
		assert not context.blockEndReached()
		parse_offset = context.getOffset()
		matchedResult = context.findNextInline(self.inlineParsers)
		# If an inline parser recognised the block content then we can parse
		# it without problem
		if matchedResult:
			# We append the text between the search start offset and the matched
			# block start
			text = context.currentFragment()[:matchedResult[0]]
			if text:
				text = textProcessor( context, text )
				text_node = context.document.createTextNode(text)
				node.appendChild(text_node)
			new_offset = matchedResult[2].parse(context, node, matchedResult[1])
			# We increase the offset so that the next parsing offset will be
			# the end of the parsed inline.
			context.increaseOffset(new_offset)
		# When we have not found any matched result, this means that we simply
		# have to append the whole block as a text node
		else:
			assert parse_offset < context.blockEndOffset
			text = textProcessor(context,
				context.documentText[parse_offset:context.blockEndOffset]
			)
			text_node = context.document.createTextNode(text)
			node.appendChild(text_node)
			# We set the end to the block end
			context.setOffset(context.blockEndOffset)
		# We make sure the parsers have actually augmented the offset
		assert context.getOffset() >= parse_offset
		return matchedResult

	def _findNextBlockSeparator( self, context ):
		"""Returns a match object that matches the next block separator, taking
		into account possible custom block objects."""
		#FIXME: Should check if the found block separator is contained in a
		#custom block or not.
		block_match = RE_BLOCK_SEPARATOR.search(context.documentText,
		context.getOffset())
		if block_match:
			local_offset = context.getOffset()
			# We look for a markup inline between the current offset and the
			# next block separator
			while local_offset<block_match.start():
				markup_match = RE_MARKUP.search(context.documentText,
					local_offset, block_match.start())
				# If we have not found a markup, we break
				if not markup_match: break
				if markup_match:
					# We have specified that markup inlines should not be searched
					# after the block separator
					local_offset, result = self._delimitXMLMarkupBlock(context, markup_match, block_match, local_offset)
					if not result is None: return result
			# We have found a block with no nested markup
			return (block_match.start(), block_match.end())
		# There was no block separator, so we reached the document end
		else:
			return (context.documentTextLength, context.documentTextLength)

	def _delimitXMLMarkupBlock( self, context, markupMatch, blockMatch, localOffset ):
		markup_match = markupMatch
		block_match  = blockMatch
		local_offset = localOffset
		assert markup_match.start()<block_match.start()
		# Case 1: Markup is a start tag
		if Markup_isStartTag(markup_match):
			# We look for the markup end inline
			offsets = context.saveOffsets()
			context.setCurrentBlock(markup_match.end(),context.documentTextLength)
			# There may be no 2nd group, so we  have to check this. Old
			# Kiwi documents may have [start:something] instead of
			# [start something]
			markup_end = None
			if markup_match.group(1):
				markup_end = self.markupParser.findEnd(
					markup_match.group(1).strip(), context)
			context.restoreOffsets(offsets)
			# If we found an end markup
			if markup_end:
				# The returned end is relative to the start markup end
				# offset (markup_end is a couple indicating the range
				# covered by the matched end markup)
				markup_end = markup_match.end() + markup_end[1]
				# If the end is greater than the block end, then we have
				# to recurse to look for a new block separator
				# after the block end
				if markup_end > block_match.start():
					offsets = context.saveOffsets()
					context.setOffset(markup_end)
					result =  self._findNextBlockSeparator(context)
					context.restoreOffsets(offsets)
					# NOTE: This is the case where we found
					# the block
					return local_offset, result
				# Otherwise we simply increase the offset, and look for
				# other possible markup inlines
				else:
					local_offset = markup_end
			# If there was not markup end, we skip the markup inline
			else:
				local_offset = markup_match.end()
		# We have a single tag, so we simply increase the offset
		else:
			local_offset = markup_match.end()
		return local_offset, None

	def _nodeHasOffsets( self, node ):
		start, end = self._nodeGetOffsets(node)
		return start != None and end != None

	def _nodeGetOffsets( self, node ):
		start = node.getAttributeNS(None, "_start") 
		end   = node.getAttributeNS(None, "_end") 
		if start == '': start = None
		if end   == '': end   = None
		if start != None: start = int(start)
		else: start = None
		if end != None: end = int(end)
		else: end = None
		return (start,end)

	def _nodeEnsureOffsets( self, node, start=None, end=None ):
		nstart, nend = self._nodeGetOffsets(node)
		if nstart is None and start != None:
			node.setAttributeNS(None, "_start", str(start))
		if nend is None and end != None:
			node.setAttributeNS(None, "_end", str(end))

	def _updateElementOffsets( self, context, node=None, counter=0, offsets=None ):
		"""This function ensures that every element has a _start and _end
		attribute indicating the bit of original data it comes from."""
		if node == None:
			node = context.document.childNodes[0]
			self._nodeEnsureOffsets(node, 0, context.documentTextLength)
		node.setAttributeNS(None, "_number", str(counter))
		# The given offsets parameter is an array with the node number and the
		# offsets. It can be used by embedders to easily access nods by offset
		if offsets != None:
			assert len(offsets) == counter, "%s != %s" % (len(offsets) , counter)
			this_offsets = [None,None]
			offsets.append(this_offsets)
		# The first step is to fill an array with the child nodes offsets
		# Each child node may or may not have an offset
		child_nodes = tuple([n for n in node.childNodes if n.nodeType == n.ELEMENT_NODE])
		nstart, nend = self._nodeGetOffsets(node)
		if child_nodes:
			self._nodeEnsureOffsets(child_nodes[0], start=nstart)
			self._nodeEnsureOffsets(child_nodes[-1], end=nend)
		child_offsets = []
		start = end = None
		for e in child_nodes:
			counter = self._updateElementOffsets(context, e, counter + 1)
			child_offsets.append(self._nodeGetOffsets(e))
		# Once this list is created, we retried the start offset of the earliest
		# child that has a start offset, same for the end offset of the latest
		# child
		child_start = None
		child_end   = None
		if child_offsets:
			i = 0
			j = len(child_offsets) - 1
			while i < len(child_offsets) and child_offsets[i][0] == None: i += 1
			while j >= 0 and child_offsets[j][1] == None: j -= 1
			if i < len(child_offsets): child_start = child_offsets[i][0]
			if j >= 0: child_end   = child_offsets[j][1]
		# We update the current node with the child offsets (this allows node
		# that have incomplete offsets to be completed)
		self._nodeEnsureOffsets(node, child_start, child_end)
		# And now we update the children offsets again (so that they actually
		# all have offsets), because we have all the information we need to
		# actually update the children offsets
		start, end = self._nodeGetOffsets(node)
		self._propagateElementOffsets(node,start,end)
		# As we now the current node offsets, we can set the real values in the
		# fiven offsets array, by simply updating the offsets value in the
		# `this_offsets` list.
		if offsets!=None:
			o = self._nodeGetOffsets(node)
			this_offsets[0] = o[0]
			this_offsets[1] = o[1]
		# And we return the number of this node
		return counter

	def _propagateElementOffsets( self, element, start=None, end=None ):
		"""Used by the _updateElementOffsets to ensure start and end offsets in
		all children and descendants."""
		#if start is None or end is None: return
		child_nodes = list([n for n in element.childNodes if n.nodeType == n.ELEMENT_NODE]) 
		self._nodeEnsureOffsets(element, start, end)
		# At first, we set the bounds properly, so that the first child node
		# start is this node start, and the last node end is this node end
		if child_nodes:
			self._nodeEnsureOffsets(child_nodes[0],  start=start)
			self._nodeEnsureOffsets(child_nodes[-1], end=end)
		# Now 
		for child in child_nodes:
			self._propagateElementOffsets(child, start=start)
			if self._nodeGetOffsets(child)[1] != None:
				_,nstart = self._nodeGetOffsets(child)
				if nstart != None: start = nstart
		child_nodes.reverse()
		for child in child_nodes:
			self._propagateElementOffsets(child,end=end)
			if self._nodeGetOffsets(child)[0] != None:
				nend,_ = self._nodeGetOffsets(child)
				if nend != None: end = nend
		# TODO: Maybe update the element offsetsd


	# TEXT PROCESSING UTILITIES________________________________________________

	def normaliseText( self, text ):
		"""Treats tabs eols and multiples spaces as single space, plus removes
		leading and trailing spaces."""
		# We do not strip the text because white spaces are important
		return RE_SPACES.sub(u" ",text)

	def expandTabs( self, text, cut=0 ):
		"""Expands the tabs in the given text, cutting the n first characters
		of each line, where n is given by the 'cut' argument.  This tabs
		expansion algorithm works better than Python line expansion
		algorithms."""
		if not text: return ""
		new_text = ""
		for line in text.split("\n"):
			start = 0
			match = 1
			new_line = ""
			while match!=None:
				match = RE_TABS.search(line, start)
				if match:
					rest  = TAB_SIZE-operator.mod(match.start(), TAB_SIZE)
					new_line += line[start:match.start()]
					# We grow the line with additional tabs
					value =  rest+(len(match.group())-1)*TAB_SIZE
					while value>0: new_line+=" " ; value-=1
					#It is important to mutate the original line
					line = new_line+line[match.end():]
					start = len(new_line)
			new_line += line[start:]
			cut_offset = min(len(new_line), cut)
			new_text += new_line[cut_offset:] + "\n"
		# We make sure that we do not add an extra empty line
		if text[-1]!=new_text[-1]:
			return new_text[:-1]
		else:
			return new_text

	@classmethod
	def getIndentation( self, text ):
		"""Returns the indentation of a string.

		Basically if a string has the first three lines at the same
		identation level, then it is indentation will be the number
		of spaces or tabs that lead the first three lines.

		Tabs have the value given by the *TAB_SIZE* variable."""
		lines = filter(lambda x:len(x)>0, string.split(text, '\n', 4))
		indent = map(self.countLeadingSpaces, lines)
		if len(lines) == 0:
			res = 0
		elif len(lines) == 1:
			res = indent[0]
		elif len(lines) == 2:
			res = indent[1]
		else: 
			res = max(max(indent[0], indent[1]), indent[2])
		return res

	@classmethod
	def countLeadingSpaces( self, text ):
		"""Returns the number of leading spaces in the given line.
		A tab will have the value given by the TAB_SIZE global."""
		count = 0
		for char in text:
			if char==u"\t":
				count += TAB_SIZE-operator.mod(count,TAB_SIZE)
			elif char==u" ":
				count+=1
			else:
				return count
		return count

	@classmethod
	def removeLeadingSpaces( self, text, maximum=None ):
		i = 0 ; count = 0
		for char in text:
			if not maximum is None and count >= maximum:
				return text[i:]
			if char==u"\t":
				count += TAB_SIZE-operator.mod(count,TAB_SIZE)
			elif char==u" ":
				count+=1
			else:
				return text[i:]
			i += 1
		return ''

	@classmethod
	def charactersToSpaces( self, text):
		"""Returns a string where all characters are converted to spaces.
		Newlines and tabs are preserved"""
		new_text = u""
		for char in text:
			if char in ("\t", "\n", " "):
				new_text += char
			else:
				new_text += " "
		return new_text

# EOF




def Upper( text ):
	return " ".join([w[0].upper() + w[1:].lower() for w in text.split()])

def escapeHTML( text ):
	text = text.replace("&", "&amp;")
	text = text.replace(">", "&gt;")
	text = text.replace("<", "&lt;")
	return text




#!/usr/bin/env python
# Encoding: iso-8859-1
# vim: tw=80 ts=4 sw=4 noet
# -----------------------------------------------------------------------------
# Project           :   Kiwi
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre                 <sebastien@type-z.org>
# License           :   Revised BSD License
# -----------------------------------------------------------------------------
# Creation date     :   19-Nov-2003
# Last mod.         :   26-Jul-2008
# -----------------------------------------------------------------------------

import os, sys, StringIO

__doc__ = """Kiwi is an advanced markup text processor, which can be used as
an embedded processor in any application. It is fast, extensible and outputs an
XML DOM."""

__version__ = "0.8.6"
__pychecker__ = "blacklist=cDomlette,cDomlettec"

import re, string, operator, getopt, codecs

# NOTE: I disabled 4Suite support, as minidom is good as it is right now
# We use 4Suite domlette
#import Ft.Xml.Domlette
#dom = Ft.Xml.Domlette.implementation
# We use minidom implementation
import xml.dom.minidom
dom = xml.dom.minidom.getDOMImplementation()

import core, kiwi2html, kiwi2lout, kiwi2twiki

FORMATS = {
	"html":kiwi2html,
#	"lout":kiwi2lout,
	"twiki":kiwi2twiki
}

#------------------------------------------------------------------------------
#
#  Command-line interface
#
#------------------------------------------------------------------------------

USAGE = u"Kiwi v."+__version__+u""",
   A flexible tool for converting plain text markup to XML and HTML.
   Kiwi can be used to easily generate documentation from plain files or to
   convert exiting Wiki markup to other formats.

   See <http://www.ivy.fr/kiwi>

Usage: kiwi [options] source [destination]

   source:
      The text file to be parsed (usually an .stx file, "-" for stdin)
   destination:
      The optional destination file (otherwise result is dumped on stdout)

Options:

   -i --input-encoding           Allows to specify the input encoding
   -o --output-encoding          Allows to specify the output encoding
   -t --tab                      The value for tabs (tabs equal N sapces).
                                 Set to 4 by default.
   -f --offsets                  Add offsets information
   -p --pretty                   Pretty prints the output XML, this should only
                                 be used for viewing the output.
   -m --html                     Outputs an HTML file corresponding to the Kiwi
                                 document
      --no-style                 Does not include the default CSS in the HTML
      --body-only                Only returns the content of the <body< element
      --level=n                  If n>0, n will transform HTML h1 to h2, etc...
   -O --output-format            Specifies and alternate output FORMAT
                                 (see below)
								 
   The available encodings are   %s
   The available formats are     %s
   
Misc:
   -h,  --help                    prints this help.
   -v,  --version                 prints the version of Kiwi.
"""

# Error codes

ERROR   = -1
INFO    = 0
SUCCESS = 1

# Normalised encodings

ASCII  = "us-ascii"
LATIN1 = "iso-8859-1"
LATIN2 = "iso-8859-2"
UTF8   = "utf-8"
UTF16  = "utf-16"
MACROMAN  = "macroman"
NORMALISED_ENCODINGS = (LATIN1, LATIN2, UTF8, UTF16, MACROMAN)

# Supported encodings

ENCODINGS = {
	ASCII:ASCII, "usascii":ASCII, "plain":ASCII, "text":ASCII, "ascii":ASCII,
	LATIN1:LATIN1, "latin-1":LATIN1, "latin1":LATIN1, "iso8859":LATIN1,
	"iso8859-1":LATIN1, "iso88591":LATIN1, "iso-88591":LATIN1,
	LATIN2:LATIN2, "latin2":LATIN2, "latin-2":LATIN2, "iso8859-2":LATIN2,
	"iso88592":LATIN2, "iso-88592":LATIN2,
	UTF8:UTF8, "utf8":UTF8,
	UTF16:UTF16, "utf16":UTF16,
	MACROMAN:MACROMAN, "mac-roman":MACROMAN
}

def run( arguments, input=None, noOutput=False ):
	"""Returns a couple (STATUS, VALUE), where status is 1 when OK, 0 when
	informative, and -1 when error, and value is a string.
	
	The given arguments can be either a string or an array, the input can be
	None (it will be then taken from the arguments), or be a file-like object,
	and the noOutput flag will not output the result on stdout or whatever file
	is given on the command line.
	"""
	if type(arguments) == str: arguments = arguments.split()

	# --We extract the arguments
	try:
		optlist, args = getopt.getopt(arguments, "hpmfO:vi:o:t:",\
		["input-encoding=", "output-encoding=", "output-format=",
		"offsets", "help", "html", "tab=", "version",
		"pretty", "no-style", "nostyle",
		"body-only", "bodyonly", "level="])
	except:
		args=[]
		optlist = []

	# We get the list of available encodings
	available_enc = []
	ENCODINGS_LIST=""
	for encoding in NORMALISED_ENCODINGS:
		try:
			codecs.lookup(encoding)
			available_enc.append(encoding)
			ENCODINGS_LIST+=encoding+", "
		except:
			pass
	ENCODINGS_LIST=ENCODINGS_LIST[:-2]+"."

	usage = USAGE % (ENCODINGS_LIST, ", ".join(FORMATS.keys()))

	# We set attributes
	pretty_print    = 0
	show_offsets    = False
	validate_output = 0
	generate_html   = 1
	no_style        = 0
	body_only       = 0
	level_offset    = 0
	input_enc       = ASCII
	output_enc      = ASCII
	output_format   = "html"
	if LATIN1 in ENCODINGS:
		input_enc  = LATIN1
		output_enc = LATIN1
	elif UTF8 in ENCODINGS:
		input_enc  = UTF8
		output_enc = UTF8

	# We parse the options
	for opt, arg in optlist:
		if opt in ('-h', '--help'):
			return (INFO, usage.encode(LATIN1))
		elif opt in ('-v', '--version'):
			return (INFO, __version__)
		elif opt in ('-i', '--input-encoding'):
			arg = string.lower(arg)
			if arg in ENCODINGS.keys() and ENCODINGS[arg] in available_enc:
				input_enc=output_enc=ENCODINGS[arg]
			else:
				r  = "Kiwi error : Specified input encoding is not available, choose between:"
				r += ENCODINGS_LIST
				return (ERROR, r)
		elif opt in ('-o', '--output-encoding'):
			arg = string.lower(arg)
			if arg in ENCODINGS.keys() and ENCODINGS[arg] in available_enc:
				output_enc=ENCODINGS[arg]
			else:
				r  = "Kiwi error: Specified output encoding is not available, choose between:"
				r += ENCODINGS_LIST
				return (ERROR, r)
		elif opt in ('-O', '--output-format'):
			arg = string.lower(arg)
			if arg in FORMATS.keys():
				output_format=arg
			else:
				r  = "Kiwi error: Given format (%s) not supported. Choose one of:\n" % (arg)
				r += "\n  - ".join(FORMATS)
				return (ERROR, r)
		elif opt in ('-t', '--tab'):
			TAB_SIZE = int(arg)
			if TAB_SIZE<1:
				return (ERROR, "Kiwi error: Specified tab value (%s) should be superior to 0." %\
				(TAB_SIZE))
		elif opt in ('--no-style', "--nostyle"):
			no_style      = 1
			generate_html = 1
			pretty_print  = 0
		elif opt in ('--body-only', "--bodyonly"):
			no_style      = 1
			body_only     = 1
			generate_html = 1
			pretty_print  = 0
		elif opt in ('-p', '--pretty'):
			pretty_print  = 1
			generate_html = 0
		elif opt in ('-m', '--html'):
			generate_html = 1
			output_format = "html"
			pretty_print  = 0
		elif opt in ('-f', '--offsets'):
			show_offsets = True
		elif opt in ('--level'):
			level_offset = min(10, max(0, int(arg)))

	# We check the arguments
	if input==None and len(args)<1:
		return (INFO, usage.encode("iso-8859-1"))

	# We set default values
	if input == None: source = args[0]
	else: source = None
	output = None
	if len(args)>1: output = args[1]

	#sys.stderr.write("Kiwi started with input as %s and output as %s.\n"\
	#% (input_enc, output_enc))
	if input: base_dir = os.getcwd()
	elif source=='-': base_dir = os.path.abspath(".")
	else: base_dir = os.path.abspath(os.path.dirname(source))

	parser = core.Parser(base_dir, input_enc, output_enc)

	if source == output and not noOutput:
		return(ERROR, "Cannot overwrite the source file.")

	# We open the input file, taking care of stdin
	if input != None:
		ifile = input
	elif source=="-":
		ifile = sys.stdin
	else:
		try:
			ifile = codecs.open(source,"r",input_enc)
		except:
			return (ERROR, "Unable to open input file: %s" % (input))

	if noOutput: pass
	elif output==None: ofile = sys.stdout
	else: ofile = open(output,"w")

	try:
		data = ifile.read()
	except UnicodeDecodeError, e:
		r  = "Impossible to decode input %s as %s\n" % (source, input_enc)
		r += "--> %s\n" % (e)
		return (ERROR, r)

	if source!="-": ifile.close()

	if type(data) != unicode:
		data = data.decode(input_enc)
	xml_document = parser.parse(data, offsets=show_offsets)

	result = None
	if generate_html:
		variables = {}
		variables["LEVEL"] = level_offset
		css_file = file(os.path.join(os.path.dirname(kiwi2html.__file__), "screen-kiwi.css"))
		if not no_style:
			variables["HEADER"] = "\n<style><!-- \n%s --></style>" % (css_file.read())
			variables["ENCODING"] = output_enc
		css_file.close()
		result = FORMATS[output_format].processor.generate(xml_document, body_only, variables)
		if result: result = result.encode(output_enc)
		else: result = ""
		if not noOutput: ofile.write(result)
	elif pretty_print:
		#Ft.Xml.Lib.Print.PrettyPrint(xml_document, ofile, output_enc)
		#MiniDom:
		result = xml_document.toprettyxml("  ").encode(output_enc)
		if not noOutput: ofile.write(result)
	else:
		#Ft.Xml.Lib.Print.Print(xml_document, ofile, output_enc)
		#MiniDom:
		result = xml_document.toxml().encode(output_enc)
		if not noOutput: ofile.write(result)
	return (SUCCESS, result)

def text2htmlbody( text, inputEncoding=None, outputEncoding=None ):
	"""Converts the given text to HTML, returning only the body."""
	s = StringIO.StringIO(text)
	command = "-m --body-only"
	if inputEncoding: command += " -i " + inputEncoding
	if outputEncoding: command += " -o " + outputEncoding
	_, text = run(command + " --", s, noOutput=True)
	s.close()
	return text

def runAsCommand():
	status, result = run(sys.argv[1:])
	if status == ERROR:
		sys.stderr.write(result + "\n")
		sys.exit(-1)
	elif status == INFO:
		sys.stdout.write(result + "\n")
		sys.exit(0)

if __name__ == "__main__":
	runAsCommand()

# EOF



# vim: ts=4
# -----------------------------------------------------------------------------
# Project           :   Tahchee                      <http://www.ivy.fr/tachee>
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre                     <sebastien@ivy.fr>
# License           :   Revised BSD License
# -----------------------------------------------------------------------------
# Creation date     :   25-Feb-2006
# Last mod.         :   13-Jul-2006
# -----------------------------------------------------------------------------

# This is a Kiwi-formatted plugin documentation. This allows to produce
# automatic documentation.
__doc__ = """

Tahchee Imaging Module
======================

The Imaging module is an easy and powerful way to easily generate bitmap images
to integrate into your pages. This module allows you to:

	- Render text using custom fonts (TTF files put in your `Fonts` dir)
	- Generate rounded-corner images, with CSS and HTML code ready for your
	  pages

"""

NAME    = "imaging"
VERSION = None
SUMMARY = "Image generation and processing (fonts, rounded corners)"

# We try to see if the PIL is present
try:
	import Image, ImageFont, ImageDraw
except:
	Image = ImageFront = ImageDraw = None

# This is a CSS file to create a class that will allow to have tables with
# rounded corners
ROUNDED_CSS = """\
table.$(cssclass) tr.upper  td.left
{ background: url('$(images)_ul.png') no-repeat top left; width : $(size)px; height: $(size)px;}
table.$(cssclass) tr.upper  td.middle
{ background: url('$(images)_um.png') repeat-x top; width : $(size)px; }
table.$(cssclass) tr.upper  td.right
{ background: url('$(images)_ur.png') no-repeat top right; width : $(size)px; height: $(size)px;}
table.$(cssclass) tr.middle td.left
{ background: url('$(images)_ml.png') repeat-y top left; width : $(size)px; }
table.$(cssclass) tr.middle td.middle
{ background: url('$(images)_mm.png'); }
table.$(cssclass) tr.middle td.right
{ background: url('$(images)_mr.png') repeat-y top right; width : $(size)px; }
table.$(cssclass) tr.lower  td.left
{ background: url('$(images)_ll.png') no-repeat bottom left; width : $(size)px; height: $(size)px;}
table.$(cssclass) tr.lower  td.middle
{ background: url('$(images)_lm.png') repeat-x bottom left; height: $(size)px;}
table.$(cssclass) tr.lower  td.right
{ background: url('$(images)_lr.png') no-repeat bottom right; width : $(size)px; height: $(size)px;}
"""
ROUNDED_HTML_START = """\
<table class="CLASS" cellpadding="0" cellspacing="0">
	<tr class="upper">
		<td class="left"> </td>
		<td class="middle"> </td>
		<td class="right"> </td>
	</tr>
	<tr class="middle">
		<td class="left"> </td>
		<td class="middle">
"""
ROUNDED_HTML_END = """\	
	</td>
		<td class="right"> </td>
	</tr>
	<tr class="lower">
		<td class="left"> </td>
		<td class="middle"> </td>
		<td class="right"> </td>
	</tr>
</table>"""

class ImagingPlugin:

	def __init__( self, site ):
		self.site = site
	
	def name( self ): return NAME
	def summary( self ): return SUMMARY
	def version( self ): return VERSION
	def doc( self ): return __doc__

	def install( self, localdict ):
		localdict["imaging"] = self

	def text( self, text, fontName, fontSize, color, folder = "images/text" ):
		"""Creates a transparent PNG image containing the given text, printed
		using the given font (taken from the fonts folder), with the given color
		(as a RGB tuple, or as a string with the hexadecimal code prefixed by
		#), and saves the image to the folder (by default, 'images/text'),
		relative to the site root. The image will be named TXXX.png, where the
		XXX is a sequence of 40 characters identifying the text image.
		
		This function returns the path to the image."""
		fontPath = self.fontsDir + "/" + fontName
		if not os.path.exists(fontPath):
			err("Font '%s' does not exist in fonts directory" % (fontName))
			return
		font  = ImageFont.truetype(fontPath, int(fontSize))
		image = Image.new("RGBA", font.getsize(text), (255,255,255,0))
		draw  = ImageDraw.Draw(image)
		draw.text((0,0), text, color, font=font)
		del draw
		output_path = self.pages() + "/" + folder
		if not os.path.exists(output_path): os.makedirs(output_path)
		sig = hashfunc.new(text + fontName + str(fontSize) + str(color))
		f = open(output_path + "/" + sig.hexdigest() + ".png", "w")
		image.save(f, "PNG")
		f.close()
		self.createdFiles.append(folder + "/" + sig.hexdigest() + ".png")
		return folder + "/" + sig.hexdigest() + ".png"

	def roundRectangleCSS(self, cssClass, imagesDir, imagesPrefix, cornerSize):
		"""Returns the definition for a table class that defines a rounded rectangle"""
		if imagesDir:
			css = ROUNDED_CSS.replace("$(images)",   imagesDir + "/" + imagesDir + "/" + imagesPrefix)
		else:
			css = ROUNDED_CSS.replace("$(images)",   imagesDir +"/" + imagesPrefix)
		css = css.replace("$(size)",     cornerSize)
		css = css.replace("$(cssclass)", cssClass)
		return css
	
	def roundRectangleHTML(self, className, start=True, end=False):
		if start:
			return ROUNDED_HTML_START.replace("CLASS", className)
		if end:
			return ROUNDED_HTML_END

# EOF



#!/usr/bin/python
# Encoding: ISO-8859-1
# vim: tw=80 ts=4 sw=4 fenc=latin-1 noet
# -----------------------------------------------------------------------------
# Project           :   Tahchee                     <http://www.ivy.fr/tahchee>
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre                     <sebastien@ivy.fr>
# License           :   Revised BSD License
# -----------------------------------------------------------------------------
# Creation date     :   20-Mar-2005
# Last mod.         :   08-May-2007
# -----------------------------------------------------------------------------

import sys ; sys.path.insert(0, "Sources")
from tahchee import main
from distutils.core import setup

SUMMARY     = "Automated static and dynamic web site creation tool"
DESCRIPTION = """\
Tahchee is a tool for developers and Web designers that makes it possible to
easily build a static Web site using the Cheetah template system. It is used to
fill in the gap between bare template and macro processing system and dynamic
template-based Web sites. It acts both as a build system (a la "make") as well
as an extension to the Cheetah template that makes it really easy to build small
to medium-sized sites. It is ideal for writing open source project or small
company Web sites.\
"""
# ------------------------------------------------------------------------------
#
# SETUP DECLARATION
#
# ------------------------------------------------------------------------------

setup(
    name        = "Tahchee",
    version     = main.__version__,
    author      = "Sebastien Pierre", author_email = "sebastien@type-z.org",
    description = SUMMARY, long_description = DESCRIPTION,
    license     = "Revised BSD License",
    keywords    = "web, Cheetah, automated, static, dynamic, build",
    url         = "http://www.ivy.fr/tahchee",
    download_url= "http://www.ivy.fr/tahchee/tahchee-%s.tar.gz" % (main.__version__) ,
    package_dir = { "": "Sources" },
    package_data= { "tahchee.plugins._kiwi": ["*.css"] },
    packages    = ["tahchee", "tahchee.plugins", "tahchee.plugins._kiwi"],
    scripts     = ["Scripts/tahchee"],
    classifiers = [
      "Development Status :: 4 - Beta",
      "Environment :: Web Environment",
      "Intended Audience :: Developers",
      "Intended Audience :: Information Technology",
      "License :: OSI Approved :: BSD License",
      "Natural Language :: English",
      "Topic :: Internet :: WWW/HTTP :: Site Management",
      "Topic :: Software Development :: Build Tools",
      "Topic :: Text Processing :: Markup",
      "Operating System :: POSIX",
      "Operating System :: Microsoft :: Windows",
      "Programming Language :: Python",
    ]
)

# EOF



# vim: ts=4
# -----------------------------------------------------------------------------
# Project           :   Tahchee                      <http://www.ivy.fr/tachee>
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre                     <sebastien@ivy.fr>
# License           :   Revised BSD License
# -----------------------------------------------------------------------------
# Creation date     :   25-Feb-2006
# Last mod.         :   02-May-2007
# -----------------------------------------------------------------------------

import os, sys

__doc__ = """

"""

NAME    = "linking"
VERSION = None
SUMMARY = "Useful functions to manage links within a site"

class LinkingPlugin:

	def __init__( self, site ):
		self.site = site

	def name( self ): return NAME
	def summary( self ): return SUMMARY
	def version( self ): return VERSION
	def doc( self ): return __doc__

	def install( self, localdict ):
		localdict["linking"] = self

	def hierarchy( self, pagePath ):
		"""Creates an HTML string that contains the clickable path from the
		website root to the current page. This can be placed in a navigation
		bar."""

		parents  = (pagePath.split("/"))
		page     = parents[-1]
		parents  = parents[:-1]

		# The first element is the site root
		res = "<a href='%s'>%s</a> / " %\
		( self.link(pagePath, "/index.html"), self.name())
		path = ""

		# Now we add links for the parents
		for path_index in range(0, len(parents)):
			path += parents[path_index] + "/"
			res += "<a href='%s'>%s</a> / " % (
				self.link(pagePath,
				path+"/index.html"),
				parents[path_index]
			)

		# We add the last link ofr the file
		radix = os.path.splitext(page)[0]
		if radix!="index":
			return res+"<a href='%s'>%s</a> " % (
				self.link(pagePath, radix), radix
			)
		else:
			# We get rid of the trailing " / "
			return res[:-3]

	def _normalize( self, path ):
		"""Normalizes the given path, fixing some issues with Windows \\ in
		paths."""
		return path.replace("\\", "/")

	def _abspath( self, path ):
		"""Ensures that the path is absolute. This does not use the Python
		os.abspath method, but simply ensures that the path does not starts with
		'.' and starts with a '/'."""
		if not path: return "/"
		path = self._normalize(path)
		if path[0] == ".": path = path[1:]
		if not path or not path[0] == "/": path = "/" + path
		return path

	def a( self, target, content ):
		return "<a href='%s'>%s</a>" % (target, content)

	def link( self, fromPath, toPath, checkLink=True ):
		"""Creates a relative or absolute link (if the site is in local mode,
		then the link is relative, otherwise it is absolute) from the given path
		to the other path. The 'fromPath' is RELATIVE TO THE PAGES DIRECTORY."""
		# WE SHOULD ASSERT THAT FROM PATH IS A FILE, OR IF IT IS A DIRECTORY, IT
		# MUST END WITH /
		fromPath = self._abspath(fromPath)
		toPath   = self._abspath(toPath)
		# Now, all paths are of the form
		# - '/'
		# - '/FILE' or '/DIR'
		# - '/DIR/FILE' or '/DIR/DIR'
		# - ...
		from_el   = fromPath.split("/")[1:]
		to_el     = toPath.split("/")[1:]
		from_dirs = len(from_el) > 1 and from_el[:-1] or []
		to_dirs   = len(to_el) > 1 and to_el[:-1] or []
		from_file = from_el[-1]
		to_file   = to_el[-1]
		common    = -1
		for c in range(0, min(len(from_dirs), len(to_dirs))):
			if from_dirs[c] != to_dirs[c]:
				break
			common = c
		# If there is no "to_file", we force the "/"
		if not to_file: to_file = "/"
		res = None
		# Both paths have the same directories in common
		if from_dirs == to_dirs:
			res = to_file
		else:
			prefix  = ""
			if common == -1:
				prefix += "../" * (len(from_dirs))
				prefix += "/".join(to_dirs)
			else:
				common += 1
				prefix += "../" * (len(from_dirs[common:]))
				if common < len(to_dirs):
					prefix +=  "/".join(to_dirs[common:])
			if prefix and not prefix[-1] == "/": prefix += "/"
			res = prefix + to_file
		return res

# EOF



#!/usr/bin/env python
# Encoding: iso-8859-1
# vim: tw=80 ts=4 sw=4 noet
# -----------------------------------------------------------------------------
# Project           :   Kiwi
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre                 <sebastien@type-z.org>
# -----------------------------------------------------------------------------
# Creation date     :   07-Feb-2006
# Last mod.         :   07-Oct-2009
# -----------------------------------------------------------------------------

import re, xml.dom
import sys
from formatting import *
import templates

#------------------------------------------------------------------------------
#
#  Processing functions
#
#------------------------------------------------------------------------------

class Processor(templates.Processor):

	def defaultProcessElement( self, element, selector ):
		"""We override this for elements with the 'html' attribute."""
		if element.getAttributeNS(None, "_html"):
			res = "<" + element.nodeName
			att = element.attributes
			for i in range(att.length):
				a = att.item(i)
				if a.name == "_html": continue
				res += " %s='%s'" % (a.name, element.getAttributeNS(None, a.name))
			if element.childNodes:
				res += ">"
				for e in element.childNodes:
					res += self.processElement(e)
				res += "</%s>" % (element.tagName)
			else:
				res += "/>"
			return res
		else:
			return templates.Processor.defaultProcessElement(self,element,selector)

	def generate( self, xmlDocument, bodyOnly=False, variables={} ):
		node = xmlDocument.getElementsByTagName("Document")[0]
		self.variables = variables
		if bodyOnly:
			for child in node.childNodes:
				if child.nodeName == "Content":
					return convertContent_bodyonly(child)
		else:
			return convertDocument(node)

#------------------------------------------------------------------------------
#
#  Actual element processing
#
#------------------------------------------------------------------------------

def convertDocument(element):
	return process(element, """\
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
<meta http-equiv="Content-Type" content="text/html; charset=$(=ENCODING)" />
$(Header)$(=HEADER)
</head>
<body>
$(Header:title)
<div class="kiwiContent">$(Content)</div>
$(References)
</body>
</html>""")

def element_number( element ):
	"""Utility function that returns the element number (part of the element
	offset attributes)"""
	number = element.getAttributeNS(None, "_number")
	if number: return int(number)
	else: return None

def wdiv( element, text ):
	"""Wraps the given text in a DIV extended with offsets attributes if the
	given element has offset attributes."""
	number = element_number(element)
	if number == None: return text
	return "<div class='KIWI N%s' ostart='%s' oend='%s'>%s</div>" % (
		element.getAttributeNS(None, '_number'),
		element.getAttributeNS(None, '_start'),
		element.getAttributeNS(None, '_end'),
		text
	)

def wspan( element, text ):
	"""Wraps the given text in a SPAN extended with offsets attributes if the
	given element has offset attributes."""
	number = element_number(element)
	if number == None: return text
	return "<div class='KIWI N%s' ostart='%s' oend='%s'>%s</div>" % (
		element.getAttributeNS(None, '_number'),
		element.getAttributeNS(None, '_start'),
		element.getAttributeNS(None, '_end'),
		text
	)

def wattrs( element ):
	"""Returns the offset attributes of this element if it has any."""
	res = ""
	number = element_number(element)
	if number != None:
		res = " class='KIWI N%s' ostart='%s' oend='%s'" % (
			element.getAttributeNS(None, '_number'),
			element.getAttributeNS(None, '_start'),
			element.getAttributeNS(None, '_end')
		)
	return res

def convertContent( element ):
	return process(element, wdiv(element, """<div class='content'>$(*)</div>"""))

def convertContent_bodyonly( element ):
	return process(element, wdiv(element, """$(*)"""))

def convertContent_table( element ):
	return process(element, """<tbody%s>$(*)</tbody>""" % (wattrs(element)))

def convertHeader( element ):
	return process(element, "<title%s>$(Title/title)</title>" % (wattrs(element)))

def convertHeading( element ):
	return process(element, wspan(element, "$(*)"))

def getSectionNumberPrefix(element):
	if not element:
		return ""
	if not element.nodeName in ("Chapter", "Section"):
		return ""
	parent = element.parentNode
	section_count = 1
	for child in parent.childNodes:
		if child == element:
			break
		if child.nodeName in ("Chapter", "Section"):
			section_count += 1
	parent_number = getSectionNumberPrefix(parent.parentNode)
	if parent_number:
		return "%s.%s" % (parent_number, section_count)
	else:
		return str(section_count)

def formatSectionNumber(number):
	number = str(number).split(".")
	depth = 0
	res   = []
	for n in number:
		if depth == len(number) - 1:
			res.append('<span class="level%s">%s<span class="lastDot dot">.</span></span>' % (depth,n))
		else:
			res.append('<span class="level%s">%s<span class="dot">.</span></span>' % (depth,n))
		depth += 1
	return "".join(res)

def convertSection( element ):
	offset = element._processor.variables.get("LEVEL") or 0
	level = int(element.getAttributeNS(None, "_depth")) + offset
	return process(element,
	  '<div class="section" level="%d">' % (level)
	  + '<h%d class="heading"><span class="number">%s</span>$(Heading)</h%d>' % (level, formatSectionNumber(getSectionNumberPrefix(element)), level)
	  + '<div class="level%d">$(Content:section)</div></div>' % (level)
	)

def convertReferences( element ):
	return process(element, """<div class="kiwiReferences">$(Entry)</div>""")

def convertEntry( element ):
	return process(element, """<div class="entry"><div class="name"><a name="%s">%s</a></div><div class="content">$(*)</div></div>""" %
	(element.getAttributeNS(None, "id"), element.getAttributeNS(None, "id")))

def convertHeader_title( element ):
	return process(element, """<div
	class="title">$(Title/title:header)$(Title/subtitle:header)</div>$(Meta)""")

def converttitle_header( element ):
	return process(element, """<h1%s>$(*)</h1>""" % (wattrs(element)))

def convertsubtitle_header( element ):
	return process(element, """<h2%s>$(*)</h2>""" % (wattrs(element)))

def convertParagraph( element ):
	return process(element, """<p%s>$(*)</p>""" % (wattrs(element)))

def convertParagraph_cell( element ):
	return process(element, """$(*)<br />""")

def convertList( element ):
	list_type = element.getAttributeNS(None, "type")
	attrs = [""]
	if list_type:
		attrs.append('class="%s"' % (list_type))
	if list_type == "ordered":
		return process(element, """<ol%s%s>$(*)</ul>""" % (wattrs(element), " ".join(attrs)))
	else:
		return process(element, """<ul%s%s>$(*)</ul>""" % (wattrs(element), " ".join(attrs)))

def convertListItem( element ):
	attrs   = [""]
	is_todo = element.getAttributeNS(None, "todo")
	if is_todo:
		if is_todo == "done":
			attrs.append('class="todo done"')
			return process(element, """<li%s%s><input type='checkbox' checked='true' readonly>$(*)</input></li>""" % (wattrs(element), " ".join(attrs)))
		else:
			attrs.append('class="todo"')
			return process(element, """<li%s%s><input type='checkbox' readonly>$(*)</input></li>""" % (wattrs(element), " ".join(attrs)))
	else:
		return process(element, """<li%s%s>$(*)</li>""" % (wattrs(element), " ".join(attrs)))

def convertTable( element ):
	tid = element.getAttributeNS(None, "id")
	if tid: tid = ' id="%s"' % (tid)
	else:   tid = ""
	return process(element, """<div class="table"%s><table cellpadding="0" cellspacing="0" align="center">$(Caption)$(Content:table)</table></div>""" % (tid))

def convertDefinition( element ):
	return process(element, """<dl%s>$(*)</dl>""" % (wattrs(element)))

def convertDefinitionItem( element ):
	return process(element, """<dt>$(Title)</dt><dd>$(Content)</dd>""")

def convertCaption( element ):
	return process(element, """<caption%s>$(*)</caption>""" % (wattrs(element)))

def convertRow( element ):
	try: index = element.parentNode.childNodes.index(element) % 2 + 1
	except: index = 0 
	classes = ( "", "even", "odd" )
	return process(element, """<tr class='%s'%s>$(*)</tr>""" % (classes[index], wattrs(element)))

def convertCell( element ):
	cell_attrs = ""
	node_type  = element.getAttributeNS(None, "type")
	if element.hasAttributeNS(None, "colspan"):
		cell_attrs += " colspan='%s'" % (element.getAttributeNS(None, "colspan"))
	if node_type == "header":
		return process(element, """<th%s%s>$(*:cell)</th>""" % (cell_attrs,wattrs(element)))
	else:
		return process(element, """<td%s%s>$(*:cell)</td>""" % (cell_attrs,wattrs(element)))

def convertBlock( element ):
	title = element.getAttributeNS(None,"title") or element.getAttributeNS(None, "type") or ""
	css_class = ""
	if title:
		css_class=" class='ann%s'" % (element.getAttributeNS(None, "type").capitalize())
		title = "<div class='title'>%s</div>"  % (title)
		div_type = "div"
	elif not element.getAttributeNS(None, "type"):
		div_type = "blockquote"
	return process(element, """<%s%s>%s<div class='content'%s>$(*)</div></%s>""" % (div_type, css_class, title, wattrs(element), div_type))

def stringToTarget( text ):
	return text.replace("  ", " ").strip().replace(" ", "_")

def convertlink( element ):
	if element.getAttributeNS(None, "type") == "ref":
		return process(element, """<a href="#%s" class="internal">$(*)</a>""" %
		(stringToTarget(element.getAttributeNS(None, "target"))))
	else:
		# TODO: Support title
		return process(element, """<a href="%s" class="external">$(*)</a>""" %
		(element.getAttributeNS(None, "target")))

def converttarget( element ):
	name = element.getAttributeNS(None, "name")
	return process(element, """<a class="anchor" name="%s">$(*)</a>""" % (stringToTarget(name)))

def convertMeta( element ):
	return process(element, "<table class='kiwiMeta'>$(*)</table>")

def convertmeta( element ):
	return process(element,
	"<tr><td width='0px' class='name'>%s</td><td width='100%%' class='value'>$(*)</td></tr>" %
	(element.getAttributeNS(None, "name")))

def convertemail( element ):
	mail = ""
	for c in  process(element, """$(*)"""):
		mail += "&#%d;" % (ord(c))
	return """<a href="mailto:%s">%s</a>""" % (mail, mail)

def converturl( element ):
	return process(element, """<a href="$(*)" target="_blank">$(*)</a>""")

def converturl_header( element ):
	return process(element, """<div class='url'>%s</div>""" % (
	converturl(element)))

def convertterm( element ):
	return process(element, """<span class='term'>$(*)</span>""")

def convertquote( element ):
	return process(element, """&ldquo;<span class='quote'>$(*)</span>&rdquo;""")

def convertcitation( element ):
	return process(element, """&laquo;<span class='citation'>$(*)</span>&raquo;""")

def convertstrong( element ):
	return process(element, """<strong>$(*)</strong>""")

def convertpre( element ):
	return process(element, """<pre%s>$(*)</pre>""" % (wattrs(element)))

def convertcode( element ):
	return process(element, """<code>$(*)</code>""")

def convertemphasis( element ):
	return process(element, """<em>$(*)</em>""")

def convertbreak( element ):
	return process(element, """<br />""")

def convertnewline( element ):
	return process(element, """<br />""")

def convertarrow( element ):
	arrow = element.getAttributeNS(None, "type")
	if   arrow == "left":
		return "&larr;"
	elif arrow == "right":
		return "&rarr;"
	else:
		return "&harr;"

def convertdots( element ):
	return "&hellip;"

def convertendash( element ):
	return "&ndash;"

def convertemdash( element ):
	return "&mdash;"

def convertentity( element ):
	return "&%s;" % (element.getAttributeNS( None, "num"))

# We create the processor, register the rules and define the process variable
processor      = Processor()
name2functions = {}
for symbol in filter(lambda x:x.startswith("convert"), dir()):
	name2functions[symbol] = eval(symbol)
processor.register(name2functions)
process = processor.process

# EOF



#!/usr/bin/env python
# Encoding: iso-8859-1
# vim: tw=80 ts=4 sw=4 noet fenc=latin-1
# -----------------------------------------------------------------------------
# Project           :   Kiwi
# -----------------------------------------------------------------------------
# Author            :   Sebastien Pierre (SPE)           <sebastien@type-z.org>
# -----------------------------------------------------------------------------
# Creation date     :   06-Mar-2006
# Last mod.         :   18-Oct-2006
# -----------------------------------------------------------------------------

import re, xml.dom
import sys
from formatting import *

RE_EXPRESSION    =re.compile("\$\(([^\)]+)\)")

__doc__ = """\
The template module implements a simple way to convert an XML document to another
format (HTML, text, whatever) by expanding very simple XPath-like expressions. Of
course, it is a very-small subset of what you could do in XSLT, but it has a
Python syntax and is very, very easy to use.

At first, your define `convertXXX' functions where XXX is your case-sensitive
tagName. All these functions should take an element as parameter and run the
`process' function with the element and template string as parameter.

Expressions like `$(XXX)' (XXX being the expression) in the template strings
will be expanded by procesing the set of nodes indicated by the XXX expression.

 - $(*) will process all children
 - $(MyNode) will process only nodes with 'MyNode' tagName
 - $(MyParent/MyNode) will process only the children of MyParent named MyNode
 - $(MyNode:alt) will use the function `convertMyNode_alt' instead of the
   default 'convertMyNode', if available. This allows to setup 'processing
   modes' for your document (like table of content, references, etc).
 - $(MyNode:alt?) will use the function `convertMyNode_alt' if it is defined, or
   fallback to `convertMyNode`.

"""

#------------------------------------------------------------------------------
#
#  Processing functions
#
#------------------------------------------------------------------------------

class Processor:
	"""The processor is the core of the template engine. You give it a Python
	module with "convert*" functions, and it will process it."""

	def __init__( self, module=None ):
		self.expressionTable = {}
		self.variables       = {}

	def register( self, name2functions ):
		"""Fills the EXPRESSION_TABLE which maps element names to processing
		functions. This function is only useful when you implement your
		templates in the same way as the `kiwi2html` module, ie. with processing
		functions like `convertXXX_VVV` where `XXX` stands for the element name,
		and `_VVV` is the optional variant (selected by`$(element:variant)`).
		
		You may prefer to use the `registerElementProcessor` instead if you want
		to register a processor for an individual tag.
		"""
		self.expressionTable = {}
		for name, function in name2functions.items():
			if name.startswith("convert"):
				ename = name[len("convert"):]
				ename = ename.replace("_", ":")
				self.expressionTable[ename] = function

	def registerElementProcessor( self, function, elementName, variant=None  ):
		"""Registers the given function to process the given element name and
		the given optional variant.
		
		Note that this will replace any previsously registered processor for the
		element and variant."""
		if variant: elementName += ":" + variant
		self.expressionTable[elementName] = function

	def resolveSet( self, element, names ):
		"""Resolves the set of names in the given element. When ["Paragraph"] is
		given, then all child paragraph nodes of the current node will be returned,
		while ["Section", "Paragraph"] will return all paragraphs for all
		sections."""
		s = []
		if len(names) == 1:
			name = names[0]
			for child in element.childNodes:
				if name != "*" and not child.nodeType == xml.dom.Node.ELEMENT_NODE: continue
				if name == "*" or child.tagName == name: s.append(child)
		else:
			name = names[0]
			for child in element.childNodes:
				if name != "*" and not child.nodeType == xml.dom.Node.ELEMENT_NODE: continue
				if name == "*" or child.tagName == name: s.extend(self.resolveSet(child, names[1:]))
		return s

	def processElement( self, element, selector=None ):
		"""Processes the given element according to the EXPRESSION_TABLE, using the
		given selector to select an alternative function."""
		selector_optional = False
		if selector and selector[-1] == "?":
			selector = selector[:-1]
			selector_optional = True
		if element.nodeType == xml.dom.Node.TEXT_NODE:
			return escapeHTML(element.data)
		elif element.nodeType == xml.dom.Node.ELEMENT_NODE:
			element._processor = self
			fname = element.nodeName
			if selector: fname += ":" + selector
			func  = self.expressionTable.get(fname)
			# There is a function for the element in the EXPRESSION TABLE
			if func:
				return func(element)
			elif selector_optional:
				self.processElement(element)
			# Otherwise we simply expand its text
			else:
				return self.defaultProcessElement(element, selector)
		else:
			return ""

	def defaultProcessElement( self, element, selector ):
		"""Default function for processing elements. This returns the text."""
		return "".join([self.processElement(e) for e in element.childNodes])

	def interpret( self, element, expression ):
		"""Interprets the given expression for the given element"""
		assert self.expressionTable
		# =VARIABLE means that we replace the expression by the content of the
		# variable in the varibales directory
		if expression.startswith("="):
			vname = expression[1:].upper()
			return self.variables.get(vname) or ""
		# Otherwise, the expression is a node selection expression, which may also
		# have a selector
		elif expression.rfind(":") != -1:
			names, selector = expression.split(":")
		# There may be no selector as well
		else:
			names           = expression
			selector        = None
		names = names.split("/")
		r     = ""
		for element in self.resolveSet(element, names):
			r += self.processElement(element, selector)
		return r

	# SYNTAX: $(EXPRESSION)
	# Where EXPRESSION is a "/" separated list of element names, optionally followed
	# by a colon ':' and a name
	def process( self, element, text ):
		i = 0
		r = ""
		while i < len(text):
			m = RE_EXPRESSION.search(text, i)
			if not m:
				r += text[i:]
				break
			else:
				r += text[i:m.start()]
			r+= self.interpret(element, m.group(1))
			i = m.end()
		return r

	def generate( self, xmlDocument, bodyOnly=False, variables={} ):
		self.variables = variables
		return self.processElement(xmlDocument.childNodes[0])

# EOF
